"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Translation of IO operation definition."""


from typing import List, Tuple

from nagini_translation.lib.constants import EVAL_IO_SIGNATURE, PRIMITIVES
from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import (
    PythonIOOperation,
    PythonVar,
)
from nagini_translation.lib.typedefs import (
    Function,
    Info,
    Method,
    Position,
    Predicate,
    VarDecl,
)
from nagini_translation.translators.io_operation.common import (
    IOOperationCommonTranslator,
)
from nagini_translation.translators.io_operation.termination_check import (
    TerminationCheckGenerator,
)
from nagini_translation.translators.io_operation.utils import (
    construct_getter_name,
)


class IOOperationDefinitionTranslator(IOOperationCommonTranslator):
    """Class responsible for translating IO operation definitions."""

    def translate_io_operation(
            self, operation: PythonIOOperation,
            ctx: Context) -> Tuple[Predicate, List[Function], List[Method]]:
        """Translate IO operation to Silver.

        This means:

        1.  Create a predicate representing IO operation.
        2.  Create body-less functions for all IO operation results.
        3.  Create a method for IO operation termination check.

        .. todo:: Vytautas

            Change from ``List`` of methods to single method, if it
            turns out that we need only one method.
        """
        args = [
            arg.decl
            for arg in operation.get_parameters()
        ]
        position = self.to_position(operation.node, ctx)
        info = self.no_info(ctx)

        predicate = self.viper.Predicate(operation.sil_name, args, None,
                                         position, info)

        getters = [
            self._construct_getter(
                operation, result, args, position, info, ctx)
            for result in operation.get_results()]

        if not operation.is_basic():
            self._translate_defining_getters(operation, ctx)

        method = self._create_termination_check(operation, ctx)
        checks = [method]

        return (
            predicate,
            getters,
            checks,
        )

    def _construct_getter(
            self, operation: PythonIOOperation, operation_result: PythonVar,
            args: List[VarDecl], position: Position,
            info: Info, ctx: Context) -> Function:
        name = construct_getter_name(operation, operation_result)
        typ = self.translate_type(operation_result.type, ctx)
        if operation_result.type.name not in PRIMITIVES:
            getter_result = self.viper.Result(typ, position, info)
            result_type_expr = self.type_check(
                getter_result, operation_result.type, position, ctx)
            posts = [result_type_expr]
        else:
            posts = []
        if (operation.name == EVAL_IO_SIGNATURE[0] and
                operation_result.name == EVAL_IO_SIGNATURE[3]):
            # Add postconditions about return type.
            getter_result = self.viper.Result(typ, position, info)
            func_param = operation._inputs[0].ref()
            # For each function for which it's called in the program, add a postcondition
            # stating the result type in that case.
            for func_arg, func_type in operation.func_args:
                result_type_expr = self.type_check(
                    getter_result, func_type, position, ctx)
                this_func = self.viper.EqCmp(func_arg, func_param, position, info)
                implication = self.viper.Implies(this_func, result_type_expr, position,
                                                 info)
                posts.append(implication)
        getter = self.viper.Function(
            name, args, typ, [], posts, None, position, info)
        return getter

    def _translate_defining_getters(
            self, main_operation: PythonIOOperation,
            ctx: Context) -> None:
        """Translate defining getter instances of existential variables."""
        assert not main_operation.is_basic()
        assert ctx.current_function is None
        ctx.current_function = main_operation

        existentials = main_operation.get_io_existentials()
        existentials.sort(key=lambda var: var.defining_order)

        for existential in existentials:
            node, result = existential.get_defining_info()
            getter = self.create_result_getter(node, result, ctx)
            existential.set_existential_ref(getter)

        ctx.current_function = None

    def _create_termination_check(
            self, operation: PythonIOOperation,
            ctx: Context) -> Method:
        """Create a termination check."""
        assert not ctx.current_function
        ctx.current_function = operation

        name = ctx.module.get_fresh_name(
            operation.sil_name + '__termination_check')
        parameters = [
            parameter.decl
            for parameter in operation.get_parameters()
        ]

        statement, termination_condition = self.translate_expr(
            operation.get_terminates(), ctx, target_type=self.viper.Bool)
        assert not statement
        statement, termination_measure = self.translate_expr(
            operation.get_termination_measure(), ctx,
            target_type=self.viper.Int)
        assert not statement

        generator = TerminationCheckGenerator(
            self, ctx, termination_condition, termination_measure,
            operation.get_termination_measure())
        if operation.is_basic():
            checks = generator.create_checks()
        else:
            checks = generator.create_checks(operation.get_body())

        info = self.no_info(ctx)
        position = self.to_position(operation.get_termination_measure(), ctx)

        body = self.translate_block(checks, position, info)
        pres = self._create_typeof_pres(operation.get_parameters(), ctx)
        ctx.current_function = None
        result = self.viper.Method(
            name=name, args=parameters, returns=[], pres=pres, posts=[],
            locals=[], body=body, position=self.no_position(ctx), info=info)

        return result

    def _create_typeof_pres(self, args,
                            ctx: Context):
        """
        Creates 'typeof' preconditions for function arguments.
        """
        pres = []
        for arg in args:
            if not arg.type.name in PRIMITIVES:
                type_check = self.get_parameter_typeof(arg, ctx)
                pres.append(type_check)
        return pres

    def get_parameter_typeof(self, param,
                             ctx: Context):
        """
        Creates an expression checking if the given parameter has its type,
        to be assumed in preconditions and/or postconditions. If possible,
        the expression is wrapped in an InhaleExhaleExpression s.t. it is
        just assumed, not checked. Generally this seems to be possible with
        types, but not with type arg numbers, because the latter encodes length
        for tuples.
        """
        no_pos = self.no_position(ctx)
        result = self.var_type_check(param.sil_name, param.type, no_pos, ctx)
        return result



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Utility functions."""


import ast

from nagini_contracts.contracts import CONTRACT_WRAPPER_FUNCS
from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import (
    PythonIOOperation,
    PythonVarBase,
)
from nagini_translation.lib.util import (
    InvalidProgramException,
)


def construct_getter_name(operation: PythonIOOperation,
                          result: PythonVarBase) -> str:
    """Utility function for constructing getter name."""
    return 'get__{0}__{1}'.format(
        operation.sil_name,
        result.sil_name,
    )


def raise_invalid_operation_use(error_type: str, node: ast.AST) -> None:
    """Raise InvalidProgramException."""
    raise InvalidProgramException(
        node,
        'invalid.io_operation_use.' + error_type,
    )


def raise_invalid_existential_var(error_type: str, node: ast.AST) -> None:
    """Raise InvalidProgramException."""
    raise InvalidProgramException(
        node,
        'invalid.io_existential_var.' + error_type,
    )


def raise_invalid_get_ghost_output(error_type: str, node: ast.AST) -> None:
    """Raise InvalidProgramException."""
    raise InvalidProgramException(
        node,
        'invalid.get_ghost_output.' + error_type,
    )


def get_parent(node: ast.expr) -> ast.expr:
    """A helper function to get a parent node."""
    # _parent is not a node field, it is added dynamically by our
    # code. That is why mypy reports an error here.
    if hasattr(node, '_parent'):
        return node._parent     # pylint: disable=protected-access
    else:
        return None


def is_top_level_assertion(node: ast.expr) -> bool:
    """Check if assertion represented by node is top level."""
    parent = get_parent(node)
    while (isinstance(parent, ast.BoolOp) and
           isinstance(parent.op, ast.And)):
        node = parent
        parent = get_parent(node)
    if (isinstance(parent, ast.Call) and
            isinstance(parent.func, ast.Name)):
        func_name = parent.func.id
        return func_name in CONTRACT_WRAPPER_FUNCS
    return False


def get_variable(var_name: str, ctx: Context) -> PythonVarBase:
    """Return variable by taking into account aliasing."""
    if var_name in ctx.var_aliases:
        var = ctx.var_aliases[var_name]
    else:
        var = ctx.actual_function.get_variable(var_name)
        assert var
    return var



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from typing import Dict, List, Union

from nagini_translation.lib.constants import BOOL_TYPE
from nagini_translation.lib.program_nodes import PythonMethod, PythonType, PythonVar
from nagini_translation.lib.typedefs import (
    Expr,
)
from nagini_translation.lib.util import (
    flatten,
    InvalidProgramException,
    UnsupportedException,
)
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.common import CommonTranslator


class AssignWrapper:
    """
    Represents an assignment of expr to a var named name, to
    be executed under conditions conds.
    """

    def __init__(self, name: str, conds: List, expr: ast.AST, node: ast.AST):
        self.name = name
        self.cond = conds
        self.expr = expr
        self.node = node
        self.names = {}
        self.var = None


class ReturnWrapper:
    """
    Represents a return of expr, to be executed under condition conds.
    """
    def __init__(self, cond: List, expr: ast.AST, node: ast.AST):
        self.cond = cond
        self.expr = expr
        self.node = node
        self.names = {}


class NotWrapper:
    """
    Represents a negation of the condition cond.
    """
    def __init__(self, cond):
        self.cond = cond


class BinOpWrapper:
    """
    Represents a binary operation to be performed on a variable;
    used to encode augmented assignments
    """
    def __init__(self, op: ast.BinOp, rhs: ast.AST):
        self.op = op
        self.rhs = rhs

Wrapper = Union[AssignWrapper, ReturnWrapper]


class PureTranslator(CommonTranslator):

    def translate_pure(self, conds: List, node: ast.AST,
                       ctx: Context) -> List[Wrapper]:
        method = 'translate_pure_' + node.__class__.__name__
        visitor = getattr(self, method, self.translate_pure_generic)
        return visitor(conds, node, ctx)

    def translate_pure_generic(self, conds: List,
                               node: ast.AST, ctx: Context) -> List[Wrapper]:
        raise UnsupportedException(node)

    def translate_pure_Expr(self, conds: List, node: ast.Expr,
                            ctx: Context) -> List[Wrapper]:
        if isinstance(node.value, ast.Str):
            # Ignore docstrings.
            return []
        raise UnsupportedException(node)

    def translate_pure_If(self, conds: List, node: ast.If,
                          ctx: Context) -> List[Wrapper]:
        """
        Translates an if-block to a list of Return- and AssignWrappers which
        contain the condition(s) introduced by the if-block.
        """
        cond = node.test
        cond_var = ctx.current_function.create_variable('cond',
            ctx.module.global_module.classes[BOOL_TYPE], self.translator)
        cond_let = AssignWrapper(cond_var.sil_name, conds, cond, node)
        then_cond = conds + [cond_var.sil_name]
        else_cond = conds + [NotWrapper(cond_var.sil_name)]
        then = [self.translate_pure(then_cond, stmt, ctx) for stmt in node.body]
        then = flatten(then)
        else_ = []
        if node.orelse:
            else_ = [self.translate_pure(else_cond, stmt, ctx) for stmt
                     in node.orelse]
            else_ = flatten(else_)
        return [cond_let] + then + else_

    def translate_pure_Return(self, conds: List, node: ast.Return,
                              ctx: Context) -> List[Wrapper]:
        """
        Translates a return statement to a ReturnWrapper
        """
        wrapper = ReturnWrapper(conds, node.value, node)
        return [wrapper]

    def translate_pure_AugAssign(self, conds: List, node: ast.AugAssign,
                                 ctx: Context) -> List[Wrapper]:
        """
        Translates an augmented assign statement to an AssignWrapper
        """
        assert isinstance(node.target, ast.Name)
        val = BinOpWrapper(node.op, node.value)
        wrapper = AssignWrapper(node.target.id, conds, val, node)
        return [wrapper]

    def translate_pure_Assign(self, conds: List, node: ast.Assign,
                              ctx: Context) -> List[Wrapper]:
        """
        Translates an assign statement to an AssignWrapper
        """
        assert len(node.targets) == 1
        assert isinstance(node.targets[0], ast.Name)
        wrapper = AssignWrapper(node.targets[0].id, conds, node.value, node)
        return [wrapper]

    def _translate_return_wrapper(self, wrapper: Wrapper, previous: Expr,
                                  function: PythonMethod,
                                  ctx: Context) -> Expr:
        info = self.no_info(ctx)
        position = self.to_position(wrapper.node, ctx)
        val = self._translate_wrapper_expr(wrapper, ctx)
        if wrapper.cond:
            cond = self._translate_condition(wrapper.cond,
                                             wrapper.names, ctx)
            if previous:
                return self.viper.CondExp(self.to_bool(cond, ctx), val,
                                          previous, position, info)
            else:
                return val
        else:
            if previous:
                raise InvalidProgramException(function.node,
                                              'function.dead.code')
            return val

    def _translate_assign_wrapper(self, wrapper: Wrapper, previous: Expr,
                                  function: PythonMethod,
                                  ctx: Context) -> Expr:
        info = self.no_info(ctx)
        position = self.to_position(wrapper.node, ctx)
        val = self.to_ref(self._translate_wrapper_expr(wrapper, ctx), ctx)
        if not previous:
            raise InvalidProgramException(function.node,
                                          'function.return.missing')
        if wrapper.cond:
            cond = self._translate_condition(wrapper.cond,
                                             wrapper.names, ctx)
            if wrapper.name in ctx.var_aliases:
                old_val = ctx.var_aliases[wrapper.name].ref()
            else:
                # Variable newly defined in conditional branch, so
                # there is no old value; the variable is not defined
                # if the condition is false.
                # Our encoding requires some value though, even
                # though that will never be used, so we take some dummy
                # value.
                zero = self.viper.IntLit(0, self.no_position(ctx),
                                         self.no_info(ctx))
                false = self.viper.FalseLit(self.no_position(ctx),
                                            self.no_info(ctx))
                null = self.viper.NullLit(self.no_position(ctx),
                                          self.no_info(ctx))
                dummies = {
                    self.viper.Int: zero,
                    self.viper.Bool: false,
                    self.viper.Ref: null
                }
                old_val = self.to_ref(dummies[wrapper.var.decl.typ()], ctx)
            new_val = self.viper.CondExp(cond, val, old_val, position,
                                         info)
            return self.viper.Let(wrapper.var.decl, new_val,
                                  previous, position, info)
        else:
            return self.viper.Let(wrapper.var.decl, val,
                                  previous, position, info)

    def _translate_wrapper_expr(self, wrapper: Wrapper,
                                ctx: Context) -> Expr:
        info = self.no_info(ctx)
        position = self.to_position(wrapper.node, ctx)
        if isinstance(wrapper.expr, BinOpWrapper):
            assert isinstance(wrapper, AssignWrapper)
            stmt, val = self.translate_expr(wrapper.expr.rhs, ctx)
            val = self.to_int(val, ctx)
            var = ctx.var_aliases[wrapper.name].ref()
            var = self.to_int(var, ctx)
            if isinstance(wrapper.expr.op, ast.Add):
                val = self.viper.Add(var, val, position, info)
            elif isinstance(wrapper.expr.op, ast.Sub):
                val = self.viper.Sub(var, val, position, info)
            elif isinstance(wrapper.expr.op, ast.Mult):
                val = self.viper.Mul(var, val, position, info)
            else:
                raise UnsupportedException(wrapper.node)
        else:
            stmt, val = self.translate_expr(wrapper.expr, ctx)
        if stmt:
            raise InvalidProgramException(wrapper.expr,
                                          'purity.violated')
        return val

    def _translate_wrapper(self, wrapper: Wrapper, previous: Expr,
                           function: PythonMethod, ctx: Context) -> Expr:
        if isinstance(wrapper, ReturnWrapper):
            return self._translate_return_wrapper(wrapper, previous,
                                                  function, ctx)
        elif isinstance(wrapper, AssignWrapper):
            return self._translate_assign_wrapper(wrapper, previous,
                                                  function, ctx)
        else:
            raise UnsupportedException(wrapper)

    def _translate_to_wrappers(self, nodes: List[ast.AST],
                               ctx: Context) -> List[Wrapper]:
        return flatten([self.translate_pure([], node, ctx)for node in nodes])

    def _collect_names(self, wrappers: List[Wrapper], function: PythonMethod):
        """
        First walk through wrappers. For every assignment, we create a new
        variable with a different name. Future references to the original
        name need to refer to the new name, so we create dicts that map old
        to new names.
        """
        previous = None
        added = {}

        for wrapper in wrappers:
            if previous:
                wrapper.names.update(previous.names)
            if added:
                wrapper.names.update(added)
            added = {}
            if isinstance(wrapper, AssignWrapper):
                name = wrapper.name
                cls = self._get_wrapper_var_type(wrapper, function)
                new_name = function.create_variable(name, cls, self.translator)
                added[name] = new_name
                wrapper.var = new_name
            previous = wrapper

    def _get_wrapper_var_type(self, wrapper: AssignWrapper,
                              function: PythonMethod) -> PythonType:
        name = wrapper.name
        cls = function.get_variable(name).type
        return cls

    def translate_exprs(self, nodes: List[ast.AST],
                        function: PythonMethod, ctx: Context) -> Expr:
        """
        Translates a list of nodes to a single (let-)expression if the nodes
        are only returns, assignments and if-blocks. First translates them to
        Assign- and ReturnWrappers with conditions derived from surrounding
        if-blocks (if any), then creates one big expression out of a list
        of wrappers.
        """
        # Translate to wrapper objects
        wrappers = self._translate_to_wrappers(nodes, ctx)
        self._collect_names(wrappers, function)

        # Second walk through wrappers, starting at the end. Translate all of
        # them into one big expression. Assigns become a let, returns just the
        # returned value, and if something happens in an if block, we put it
        assert not ctx.var_aliases
        previous = None
        for wrapper in reversed(wrappers):
            ctx.var_aliases = wrapper.names.copy()
            previous = self._translate_wrapper(wrapper, previous, function, ctx)

        ctx.var_aliases = {}
        return previous

    def _translate_condition(self, conds: List, names: Dict[str, PythonVar],
                             ctx: Context) -> Expr:
        """
        Translates the conditions in conds to a big conjunctive expression,
        using the renamings in names.
        """
        previous = self.viper.TrueLit(self.no_position(ctx), self.no_info(ctx))
        for cond in conds:
            if isinstance(cond, NotWrapper):
                current = self.to_bool(ctx.var_aliases.get(cond.cond).ref(), ctx)
                current = self.viper.Not(current, self.no_position(ctx),
                                         self.no_info(ctx))
            else:
                current = ctx.var_aliases.get(cond).ref()
            previous = self.viper.And(self.to_bool(previous, ctx),
                                      self.to_bool(current, ctx),
                                      self.no_position(ctx),
                                      self.no_info(ctx))
        return previous



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Helper classes for constructing Silver AST."""

# pragma pylint: disable=wildcard-import


from nagini_translation.lib.silver_nodes.base import *
from nagini_translation.lib.silver_nodes.bool_expr import *
from nagini_translation.lib.silver_nodes.call import *
from nagini_translation.lib.silver_nodes.expression import *
from nagini_translation.lib.silver_nodes.int_cmp_expr import *
from nagini_translation.lib.silver_nodes.int_expr import *
from nagini_translation.lib.silver_nodes.location_expr import *
from nagini_translation.lib.silver_nodes.perm_cmp_expr import *
from nagini_translation.lib.silver_nodes.perm_expr import *
from nagini_translation.lib.silver_nodes.program import *
from nagini_translation.lib.silver_nodes.reference_expr import *
from nagini_translation.lib.silver_nodes.statement import *
from nagini_translation.lib.silver_nodes.types import *



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Visitors for collecting conditions that guard AST nodes."""


import ast


class GuardCollectingVisitor(ast.NodeVisitor):
    """Visitor that maintains guard information.

    For each assertion, a list of conditions that have to be true for
    that assertion to be enabled is stored in ``current_guard``
    property.
    """

    def __init__(self) -> None:
        super().__init__()
        self.current_guard = None   # type: Optional[List[ast.AST]]

    def traverse(self, node: ast.AST) -> None:
        """Traverse ``node``."""
        assert self.current_guard is None
        self.current_guard = []
        self.visit(node)
        assert self.current_guard is not None
        assert len(self.current_guard) == 0
        self.current_guard = None

    def visit_Call(self, node: ast.Call) -> None:
        """Add ``Implies`` condition to guard."""
        if (isinstance(node.func, ast.Name) and
                node.func.id == 'Implies'):
            assert len(node.args) == 2
            self.current_guard.append(node.args[0])
            self.visit(node.args[1])
            self.current_guard.pop()
        else:
            self.generic_visit(node)

    def visit_IfExp(self, node: ast.IfExp) -> None:
        """Add ``If`` condition to guard."""
        # Visit then branch.
        self.current_guard.append(node.test)
        self.visit(node.body)
        self.current_guard.pop()
        # Visit else branch.
        condition = ast.UnaryOp(ast.Not(), node.test)
        self.current_guard.append(condition)
        self.visit(node.orelse)
        self.current_guard.pop()



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Silver expressions for denoting memory locations."""


from nagini_translation.lib.silver_nodes.expression import Expression
from nagini_translation.lib.silver_nodes.bool_expr import BoolExpression
from nagini_translation.lib.silver_nodes.perm_expr import (
    FullPerm,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
)


class Location(Expression):   # pylint: disable=abstract-method
    """Denotes an access to specific location."""


class PredicateAccess(Location):
    """A predicate with one ``Ref`` argument access."""

    def __init__(self, name: str, reference: 'RefExpression') -> None:
        self._name = name
        self._reference = reference

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        reference = self._reference.translate(translator, ctx, position, info)
        return translator.viper.PredicateAccess(
            [reference], self._name, position, info)


class FieldAccess(Location):
    """Field access."""

    def __init__(self, target: 'RefExpression', field_name: str,
                 field_type: 'Type') -> None:
        self._target = target
        self._field_name = field_name
        self._field_type = field_type

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        target = self._target.translate(translator, ctx, position, info)
        field = translator.viper.Field(
            self._field_name, self._field_type.translate(translator),
            position, info)
        return translator.viper.FieldAccess(
            target, field, position, info)


class Acc(BoolExpression):
    """Access to specific location."""

    def __init__(
            self, location: Location,
            perm: 'PermExpression' = None) -> None:
        self._location = location
        self._perm = perm or FullPerm()

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        location = self._location.translate(translator, ctx, position, info)
        perm = self._perm.translate(translator, ctx, position, info)
        if isinstance(self._location, PredicateAccess):
            return translator.viper.PredicateAccessPredicate(
                location, perm, position, info)
        else:
            return translator.viper.FieldAccessPredicate(
                location, perm, position, info)


__all__ = (
    'Acc',
    'FieldAccess',
    'Location',
    'PredicateAccess',
)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Code for constructing Silver Method nodes with obligation stuff."""


import ast
from typing import List, Optional

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.config import obligation_config
from nagini_translation.lib.context import Context
from nagini_translation.lib.errors import rules
from nagini_translation.lib.program_nodes import (
    PythonMethod,
    PythonVar,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Method,
    Position,
    Stmt,
    VarDecl,
)
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.sif.lib.viper_ast_extended import ViperASTExtended
from nagini_translation.translators.obligation.manager import (
    ObligationManager,
)
from nagini_translation.translators.obligation.obligation_info import (
    PythonMethodObligationInfo,
)
from nagini_translation.translators.obligation.utils import (
    bound_obligations,
)


class ObligationMethod:
    """Info for generating Silver ``Method`` AST node."""

    def __init__(
            self, name: str, args: List[VarDecl], returns: List[VarDecl],
            pres: List[Expr], posts: List[Expr], local_vars: List[VarDecl],
            body: List[Stmt]) -> None:
        self.name = name
        self.args = args
        self.returns = returns
        self.pres = pres
        self.posts = posts
        self.local_vars = local_vars
        self.body = body

    def prepend_arg(self, arg: VarDecl, viper: Optional[ViperASTExtended] = None) -> None:
        """Prepend ``arg`` to the argument list."""
        if viper:
            info = viper.SIFInfo([], obligation_var=True)
            arg = viper.LocalVarDecl(arg.name(), arg.typ(), arg.pos(), info)
        self.args.insert(0, arg)

    def prepend_return(self, arg: VarDecl, viper: Optional[ViperASTExtended]) -> None:
        """Prepend ``arg`` to the return list."""
        if viper:
            info = viper.SIFInfo([], obligation_var=True)
            arg = viper.LocalVarDecl(arg.name(), arg.typ(), arg.pos(), info)
        self.returns.insert(0, arg)

    def prepend_body(self, statements: List[Stmt]) -> None:
        """Prepend ``statements`` to body."""
        if self.body is not None:
            self.body[0:0] = statements

    def prepend_precondition(self, preconditions: List[Expr]) -> None:
        """Prepend ``preconditions`` to precondition list."""
        self.pres[0:0] = preconditions

    def append_preconditions(self, preconditions: List[Expr]) -> None:
        """Append ``preconditions`` to precondition list."""
        self.pres.extend(preconditions)

    def append_postconditions(self, postconditions: List[Expr]) -> None:
        """Append ``postconditions`` to postcondition list."""
        self.posts.extend(postconditions)

    def prepend_postcondition(self, postcondition: Expr) -> None:
        """Prepend ``postcondition`` to postcondition list."""
        self.posts.insert(0, postcondition)

    def add_local(self, var: PythonVar) -> None:
        """Add local variable to variables list."""
        self.local_vars.append(var.decl)


class ObligationMethodNodeConstructor:
    """A class that creates a method node with obligation stuff."""

    def __init__(
            self, obligation_method: ObligationMethod,
            python_method: PythonMethod, translator: 'AbstractTranslator',
            ctx: Context, obligation_manager: ObligationManager,
            position: Position, info: Info, overriding_check: bool) -> None:
        self._obligation_method = obligation_method
        self._python_method = python_method
        self._translator = translator
        self._ctx = ctx
        self._obligation_manager = obligation_manager
        self._position = position
        self._info = info
        self._overriding_check = overriding_check
        """Are we translating a behavioral subtyping check?"""

    def construct_node(self) -> Method:
        """Construct a Silver node that represents a method."""
        method = self._obligation_method
        body = method.body
        if self._is_body_native_silver():
            # Axiomatized method, do nothing with body.
            body_block = body
        else:
            # Convert body to Scala.
            body_block = self._translator.translate_block(
                body, self._position, self._info)
        res = self._viper.Method(
            method.name, method.args, method.returns,
            method.pres, method.posts, method.local_vars, body_block,
            self._position, self._info)
        return res

    def add_obligations(self) -> None:
        """Add obligation stuff to Method."""
        self._add_additional_parameters()
        self._add_additional_returns()
        self._add_additional_variables()
        self._add_additional_preconditions()
        self._add_additional_postconditions()
        if not self._need_skip_body():
            self._set_up_measures()
            self._bound_obligations()
            self._add_body_leak_check()
            self._obligation_method.append_postconditions(
                self._obligation_info.get_additional_postconditions())
        self._add_caller_leak_check()

    def _is_body_native_silver(self) -> bool:
        """Check if body is already in Silver."""
        return (self._obligation_method.body is None or isinstance(
            self._obligation_method.body,
            self._translator.jvm.viper.silver.ast.Seqn))

    def _need_skip_body(self) -> bool:
        """Check if altering body should not be done."""
        return (self._is_body_native_silver() or
                (self._python_method.contract_only and
                 not self._overriding_check))

    def _add_additional_parameters(self) -> None:
        """Add current thread, caller measures, and residue parameters."""
        if obligation_config.disable_all:
            return
        viper = self._viper if isinstance(self._viper, ViperASTExtended) else None
        self._obligation_method.prepend_arg(
            self._obligation_info.residue_level.decl)
        if not obligation_config.disable_measures:
            self._obligation_method.prepend_arg(
                self._obligation_info.caller_measure_map.get_var().decl, viper)
        self._obligation_method.prepend_arg(
            self._obligation_info.current_thread_var.decl, viper)

    def _add_additional_returns(self) -> None:
        """Add current wait level ghost return."""
        if obligation_config.disable_waitlevel_check:
            return
        viper = self._viper if isinstance(self._viper, ViperASTExtended) else None
        self._obligation_method.prepend_return(
            self._obligation_info.current_wait_level.decl, viper)

    def _add_additional_variables(self) -> None:
        """Add current wait level ghost target."""
        if obligation_config.disable_waitlevel_check:
            return
        self._obligation_method.add_local(
            self._obligation_info.current_wait_level_target)

    def _add_additional_preconditions(self) -> None:
        """Add preconditions about current thread and caller measures."""
        preconditions = []
        if not obligation_config.disable_all:
            cthread_var = self._obligation_info.current_thread_var
            cthread = sil.RefVar(cthread_var)
            preconditions.append(cthread != None)
        position = self._position
        if (self._is_body_native_silver() and
                not obligation_config.disable_termination_check):
            # Add obligations described in interface_dict.
            assert self._python_method.interface
            for obligation in self._obligation_manager.obligations:
                preconditions.extend(
                    obligation.generate_axiomatized_preconditions(
                        self._obligation_info,
                        self._python_method.interface_dict))
            func_node = self.get_function_node()
            position = self._translator.to_position(func_node, self._ctx,
                rules=rules.OBLIGATION_CALL_LEAK_CHECK_FAIL)
        translated = [
            precondition.translate(
                self._translator, self._ctx, position, self._info)
            for precondition in preconditions]
        if not obligation_config.disable_all:
            translated.append(self._translator.var_type_check(
                cthread_var.sil_name, cthread_var.type, self._position,
                self._ctx))
        self._obligation_method.prepend_precondition(translated)
        self._obligation_method.append_preconditions(
            self._obligation_info.get_additional_preconditions())

    def _add_additional_postconditions(self) -> None:
        """Initialize current wait level for the caller."""
        if obligation_config.disable_waitlevel_check:
            return
        postcondition = self._translator.initialize_current_wait_level(
            sil.PermVar(self._obligation_info.current_wait_level),
            sil.PermVar(self._obligation_info.residue_level),
            self._ctx)
        translated = postcondition.translate(
            self._translator, self._ctx, self._position, self._info)
        self._obligation_method.prepend_postcondition(translated)

    def _set_up_measures(self) -> None:
        """Create and initialize method's measure map."""
        if obligation_config.disable_measures:
            return
        instances = self._obligation_info.get_all_precondition_instances()
        statements = self._obligation_info.method_measure_map.initialize(
            instances, self._translator, self._ctx, self._overriding_check)
        self._obligation_method.prepend_body(statements)
        self._obligation_method.add_local(
            self._obligation_info.method_measure_map.get_var())

    def _bound_obligations(self) -> None:
        """Convert all unbounded obligations to bounded ones."""
        if self._overriding_check:
            return
        statements = bound_obligations(
            self._obligation_info.get_all_precondition_instances(),
            self._translator, self._ctx, self._position, self._info)
        self._obligation_method.prepend_body(statements)

    def _add_body_leak_check(self) -> None:
        """Add a leak check.

        Check that method body does not leak obligations.
        """
        if obligation_config.disable_method_body_leak_check:
            return
        reference_name = self._python_method.get_fresh_name('_r')
        check = sil.InhaleExhale(
            sil.TrueLit(),
            self._obligation_manager.create_leak_check(reference_name))
        node = self._python_method.node
        assert node
        position = self._translator.to_position(
            node, self._ctx, rules=rules.OBLIGATION_BODY_LEAK_CHECK_FAIL)
        info = self._translator.to_info(["Body leak check."], self._ctx)
        postcondition = check.translate(
            self._translator, self._ctx, position, info)
        self._obligation_method.append_postconditions([postcondition])

    def _add_caller_leak_check(self) -> None:
        """Add a leak check.

        Check that if callee is not terminating, caller has no
        obligations.
        """
        if obligation_config.disable_call_context_leak_check:
            return
        # MustTerminate leak check.
        must_terminate = self._obligation_manager.must_terminate_obligation
        cthread = self._obligation_info.current_thread_var
        predicate = must_terminate.create_predicate_access(cthread)
        termination_leak_check = sil.CurrentPerm(predicate) == sil.NoPerm()
        # Other obligations leak check.
        reference_name = self._python_method.get_fresh_name('_r')
        leak_check = self._obligation_manager.create_leak_check(
            reference_name)
        if self._python_method.interface:
            if must_terminate.is_interface_method_terminating(
                    self._python_method.interface_dict):
                exhale = self._obligation_info.caller_measure_map.check(
                    sil.RefVar(self._obligation_info.current_thread_var),
                    sil.RawIntExpression(1))
            else:
                exhale = sil.BigAnd([termination_leak_check, leak_check])
        else:
            # Termination condition.
            tcond = self._obligation_info.create_termination_check(False)
            # Combination.
            exhale = sil.BigOr([
                tcond,
                sil.BigAnd([termination_leak_check, leak_check])
            ])
        check = sil.InhaleExhale(sil.TrueLit(), exhale)
        # Translate to Silver.
        node = self.get_function_node()
        position = self._translator.to_position(
            node, self._ctx, rules=rules.OBLIGATION_CALL_LEAK_CHECK_FAIL)
        info = self._translator.to_info(["Caller side leak check"], self._ctx)
        precondition = check.translate(
            self._translator, self._ctx, position, info)
        self._obligation_method.append_preconditions([precondition])

    def get_function_node(self):
        """
        Returns a Python AST node representing the method. If there is no existing one
        (which is the case for interface methods defined directly in Silver), creates
        a dummy node with the correct name.
        """
        if self._python_method.node is None:
            node = ast.FunctionDef()
            if self._python_method.interface_name is not None:
                node.name = self._python_method.interface_name
            else:
                node.name = self._python_method.name
            node.lineno = 0
            node.col_offset = 0
        else:
            node = self._python_method.node
        return node

    @property
    def _obligation_info(self) -> PythonMethodObligationInfo:
        return self._python_method.obligation_info

    @property
    def _viper(self) -> ViperAST:
        return self._translator.viper



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Singleton for managing obligations."""


from typing import List

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.typedefs import (
    Predicate,
)
from nagini_translation.translators.common import CommonTranslator
from nagini_translation.translators.obligation.types.base import (
    Obligation,
)
from nagini_translation.translators.obligation.types.must_terminate import (
    MustTerminateObligation,
)
from nagini_translation.translators.obligation.types.must_invoke import (
    MustInvokeObligation,
)
from nagini_translation.translators.obligation.types.must_release import (
    MustReleaseObligation,
)


class ObligationManager:
    """Class that knows about all obligation types."""

    def __init__(self) -> None:
        self._must_terminate_obligation = MustTerminateObligation()
        self._must_invoke_obligation = MustInvokeObligation()
        self._must_release_obligation = MustReleaseObligation()
        self._obligations = [
            self._must_terminate_obligation,
            self._must_invoke_obligation,
            self._must_release_obligation,
        ]

    @property
    def must_terminate_obligation(self) -> MustTerminateObligation:
        """Get ``MustTerminate`` obligation."""
        return self._must_terminate_obligation

    @property
    def must_invoke_obligation(self) -> MustInvokeObligation:
        """Get ``MustInvoke`` obligation."""
        return self._must_invoke_obligation

    @property
    def must_release_obligation(self) -> MustReleaseObligation:
        """Get ``MustRelease`` obligation."""
        return self._must_release_obligation

    @property
    def obligations(self) -> List[Obligation]:
        """Get a list of all obligations."""
        return self._obligations

    def create_predicates(
            self, translator: CommonTranslator) -> List[Predicate]:
        """Get all predicates that are used to represent obligations."""
        predicates = []
        for obligation in self._obligations:
            predicates.extend(obligation.create_predicates(translator))
        return predicates

    def create_fields(
            self, translator: CommonTranslator) -> List[Predicate]:
        """Get all fields that are used to represent obligations."""
        fields = []
        for obligation in self._obligations:
            fields.extend(obligation.create_fields(translator))
        return fields

    def create_leak_check(
            self, var_name: str) -> sil.BoolExpression:
        """Create a leak check for all obligation except termination."""
        checks = []
        for obligation in self._obligations:
            if obligation is self._must_terminate_obligation:
                continue
            checks.extend(obligation.create_leak_check(var_name))
        return sil.BigAnd(checks)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

from contextlib import contextmanager
from nagini_translation.lib.constants import ERROR_NAME, RESULT_NAME
from nagini_translation.lib.io_context import IOOpenContext
from nagini_translation.lib.obligation_context import ObligationContext
from nagini_translation.lib.program_nodes import (
    PythonMethod,
    PythonType,
    PythonVar,
    PythonVarBase,
)
from nagini_translation.lib.typedefs import Expr
from typing import Dict, List


class Context:
    """
    Contains the current state of the entire translation process.
    """

    def __init__(self) -> None:
        self.current_function = None
        self.current_class = None
        self.current_contract_exception = None
        self.var_aliases = {}
        self.old_aliases = {}
        self.label_aliases = {}
        self.position = []
        self.info = None
        self.module = None
        self.inlined_calls = []
        self.ignore_family_folds = False
        self.added_handlers = []
        self.loop_iterators = {}
        self.io_open_context = IOOpenContext()
        self.obligation_context = ObligationContext()
        self._alias_context_stack = []
        self._current_alias_context = []
        self.bound_type_vars = {}
        self._global_counter = 0
        self.perm_factor = None     # If this is set, all translated permission amounts
                                    # are multiplied by this factor.
        self._old_aliases = {}      # Keys are pretty-printed Python expressions,
                                    # values are Silver expressions they should be
                                    # translated to.
        self.ignore_waitlevel_constraints = False      # If this is set, all encountered
                                                       # WaitLevel() < e constraints will
                                                       # be translated to true.
        # Whether Abstract Read Permissions may be used
        self.arp = False
        # Stores the thread object for which the contracts are currently translated.
        self.current_thread_object = None
        # True iff contracts for a thread start are translated.
        # Used to differentiate fresh token from old token permission.
        self.is_thread_start = False

    def get_fresh_int(self) -> int:
        """
        Returns a fresh integer value, to be used as a globally used counter
        where needed. Current use case is as an argument to constructors for
        various data types, to make the instances unique.
        """
        result = self._global_counter
        self._global_counter += 1
        return result

    @property
    def all_vars(self) -> List[PythonVar]:
        """
        Returns all variables accessible in the current context, i.e., global
        variables as well as local variables and arguments of the current
        function, if any.
        """
        res = []
        if self.current_function:
            res += list(self.current_function.locals.items())
            res += list(self.current_function.args.items())
        if self.module:
            res += list(self.module.global_vars.items())
        return res

    @property
    def actual_function(self) -> PythonMethod:
        """
        Returns the function/method which is actually currently being
        translated, i.e. if a function is currently being inlined, that
        function and not the one it is being inlined into.
        """
        if not self.inlined_calls:
            return self.current_function
        return self.inlined_calls[-1]

    @property
    def result_var(self) -> PythonVar:
        """
        Returns the result var of the current function or the current alias for
        it.
        """
        if RESULT_NAME in self.var_aliases:
            return self.var_aliases[RESULT_NAME]
        if self.current_function.result:
            return self.current_function.result
        return None

    @property
    def error_var(self) -> PythonVar:
        """
        Returns the error var of the current function or the current alias for
        it.
        """
        if ERROR_NAME in self.var_aliases:
            return self.var_aliases[ERROR_NAME]
        return self.current_function.error_var

    def get_label_name(self, name: str) -> str:
        """
        Returns the actual name of the given label in the current context, i.e.,
        looks for aliases.
        """
        if name in self.label_aliases:
            return self.label_aliases[name]
        return name

    @contextmanager
    def additional_aliases(self, aliases: Dict[str, PythonVarBase]) -> None:
        """
        Execute in a context with additional aliases.
        """
        for name, var in aliases.items():
            self.set_alias(name, var, None)
        try:
            yield
        finally:
            for name in aliases:
                self.remove_alias(name)

    @contextmanager
    def aliases_context(self) -> None:
        """
        All aliases added in this context are automatically removed at
        the end of it.
        """
        self._alias_context_stack.append(self._current_alias_context)
        self._current_alias_context = []
        try:
            yield
        finally:
            for name in self._current_alias_context:
                self.remove_alias(name)
            self._current_alias_context = self._alias_context_stack.pop()

    def set_alias(self, name: str, var: PythonVar,
                  replaces: PythonVar=None) -> None:
        """
        Sets an alias for a variable. Makes sure the alt_types of the alias
        variable match those of the variable it replaces. If there already is
        an alias for the given name, memorizes the old one and replaces it.
        """
        if name in self.var_aliases:
            if name not in self.old_aliases:
                self.old_aliases[name] = []
            self.old_aliases[name].append(self.var_aliases[name])
        if replaces:
            if replaces.alt_types:
                assert not var.alt_types
                var.alt_types = replaces.alt_types
        self.var_aliases[name] = var
        self._current_alias_context.append(name)

    def remove_alias(self, name: str) -> None:
        """
        Removes the alias for the given variable. If there was a different alias
        before, that one will be used again afterwards. Otherwise, there will
        no longer be an alias for this name.
        """
        if name in self.old_aliases and self.old_aliases[name]:
            old = self.old_aliases[name].pop()
            self.var_aliases[name] = old
        elif name in self.var_aliases:
            del self.var_aliases[name]

    def set_old_expr_alias(self, key: str, val: Expr) -> None:
        self._old_aliases[key] = val

    def clear_old_expr_aliases(self) -> None:
        self._old_aliases.clear()

    @property
    def old_expr_aliases(self):
        return self._old_aliases

    def get_contents(self, only_top: bool) -> Dict:
        """
        Returns the elements that can be accessed from this container (to be
        used by get_target). If 'only_top' is true, returns only top level
        elements that can be accessed without a receiver.
        """
        return self.var_aliases




"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast
from collections import OrderedDict
from typing import List, Set, Tuple

from nagini_translation.lib.constants import (
    ARBITRARY_BOOL_FUNC,
    ASSERTING_FUNC,
    CHECK_DEFINED_FUNC,
    ERROR_NAME,
    FUNCTION_DOMAIN_NAME,
    GET_ARG_FUNC,
    GET_METHOD_FUNC,
    GET_OLD_FUNC,
    GLOBAL_CHECK_DEFINED_FUNC,
    GLOBAL_VAR_FIELD,
    IS_DEFINED_FUNC,
    JOINABLE_FUNC,
    MAY_SET_PRED,
    METHOD_ID_DOMAIN,
    PRIMITIVES,
    RESULT_NAME,
    STRING_TYPE,
    THREAD_DOMAIN,
    THREAD_POST_PRED,
    THREAD_START_PRED,
)
from nagini_translation.lib.program_nodes import (
    MethodType,
    PythonClass,
    PythonField,
    PythonMethod,
    PythonModule,
    PythonNode,
    PythonVar,
)
from nagini_translation.lib.typedefs import (
    Domain,
    DomainAxiom,
    DomainFunc,
    DomainType,
    Expr,
    Field,
    Function,
    Info,
    Method,
    Position,
    Predicate,
    Program,
    Stmt,
    Var,
    VarDecl,
)
from nagini_translation.lib.util import (
    InvalidProgramException,
)
from nagini_translation.sif.lib.viper_ast_extended import ViperASTExtended
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.common import CommonTranslator


class ProgramTranslator(CommonTranslator):
    def __init__(self, config: 'TranslatorConfig', jvm: 'JVM', source_file: str,
                 type_info: 'TypeInfo', viper_ast: 'ViperAST') -> None:
        super().__init__(config, jvm, source_file, type_info, viper_ast)
        self.required_names = {}

    def translate_field(self, field: PythonField,
                        ctx: Context) -> 'silver.ast.Field':
        return self.viper.Field(field.sil_name,
                                self.translate_type(field.type, ctx),
                                self.to_position(field.node, ctx),
                                self.no_info(ctx))

    def _translate_fields(self, cls: PythonClass,
                          ctx: Context) -> Tuple[List['silver.ast.Field'],
                                                 List['silver.ast.Function'],
                                                 List['silver.ast.Method']]:
        """
        Translates fields and properties to Viper. Normal fields get translated to
        Viper fields, properties to functions and property setters to methods.
        """
        fields = []
        functions = []
        methods = []
        for field in cls.fields.values():
            if isinstance(field, PythonField) and field.inherited is None:
                sil_field = self.translate_field(field, ctx)
                field.sil_field = sil_field
                fields.append(sil_field)
            elif isinstance(field, PythonMethod):
                # This is a property
                if field.overrides:
                    raise InvalidProgramException(field.node, 'invalid.override')
                getter = self.translate_function(field, ctx)
                functions.append(getter)
                if field.setter:
                    setter = self.translate_method(field.setter, ctx)
                    methods.append(setter)

        return fields, functions, methods

    def create_static_field_function(self, root: PythonVar,
                                     classes: List[PythonClass],
                                     ctx: Context) -> 'silver.ast.Function':
        """
        Creates a function which represents a static field. The function takes
        a parameter which represents the class on which it is called, and has
        postconditions defining its return value based on this parameter.

        'root' must be the version of the field in the class that is highest in
        the inheritance hierarchy. 'classes' must be a list of classes that
        inherit or redefine it.
        """
        current_module = ctx.module
        type = self.translate_type(root.type, ctx)
        position = self.to_position(root.node, ctx)
        info = self.no_info(ctx)
        posts = []
        result = self.viper.Result(type, position, info)
        type_type = self.type_factory.type_type()
        type_decl = self.viper.LocalVarDecl('receiver', type_type, position,
                                            info)
        type_ref = self.viper.LocalVar('receiver', type_type, position, info)
        if root.type.name not in PRIMITIVES:
            posts.append(self.type_check(result, root.type, position, ctx))
        # Iterate through all classes that 'inherit' or redefine this field
        for cls in classes:
            # Get their version (might be redefined or inherited).
            field = cls.get_static_field(root.name)
            ctx.current_class = field.cls
            ctx.module = field.cls.module
            # Compute the field value
            stmt, value = self.translate_expr(field.value, ctx)
            if stmt:
                raise InvalidProgramException('purity.violated', field.node)
            field_position = self.to_position(field.node, ctx)
            # Create a postcondition of the form
            # receiver == cls ==> result == value
            has_value = self.viper.EqCmp(result, value, field_position, info)
            type_literal = self.type_factory.translate_type_literal(
                field.cls, field_position, ctx)
            exact_type = self.viper.EqCmp(type_ref, type_literal, position,
                                          info)
            posts.append(self.viper.Implies(exact_type, has_value,
                                            field_position, info))
        ctx.module = current_module
        # Create a single function that represents all
        return self.viper.Function(root.sil_name, [type_decl], type, [], posts,
                                   None, position, info)

    def create_global_var_function(self, var: PythonVar,
                                   ctx: Context) -> 'silver.ast.Function':
        """
        Creates a Viper function representing the given global variable.
        """
        type = self.translate_type(var.type, ctx)
        position = self.to_position(var.node, ctx)
        posts = []
        result = self.viper.Result(type, position, self.no_info(ctx))
        if var.is_final:
            if var.type.name not in PRIMITIVES:
                posts.append(self.type_check(result, var.type, position, ctx))
            if hasattr(var, 'value'):
                body = None
                try:
                    stmt, value = self.translate_expr(var.value, ctx)
                    if not stmt:
                        if not self.viper.is_heap_dependent(value):
                            body = value
                            posts.append(self.viper.EqCmp(result, value, position,
                                                          self.no_info(ctx)))
                except AttributeError as e:
                    # The translation (probably) tried to access ctx.current_function
                    pass
            else:
                body = None
        else:
            body = None
        return self.viper.Function(var.sil_name, [], type, [], posts, body,
                                   self.to_position(var.node, ctx),
                                   self.no_info(ctx))

    def _create_inherit_check_postamble(self, stmts: List[Stmt], end_lbl: 'silver.ast.Label',
                                        ctx: Context) -> None:
        goto_end = self.viper.Goto(end_lbl.name(), self.no_position(ctx),
                                   self.no_info(ctx))
        stmts.append(goto_end)
        stmts += self.add_handlers_for_inlines(ctx)

        stmts.append(end_lbl)

    def create_inherit_check(self, method: PythonMethod, cls: PythonClass,
                             ctx: Context) -> 'silver.ast.Callable':
        """
        Creates a Viper function/method with the contract of the overridden
        function which calls the overriding function, to check behavioural
        subtyping.
        """
        old_function = ctx.current_function
        ctx.current_function = method
        pos = self.viper.to_position(cls.node, ctx.position)
        ctx.position.append(('inheritance', pos))
        self.info = self.viper.SimpleInfo(['behavioural.subtyping'])

        args = []
        params = []

        for arg_name, arg in method.args.items():
            args.append(arg)
            params.append(arg.decl)

        self.bind_type_vars(method, ctx)

        results = []
        locals_before = set(method.locals.values())
        if method.type:
            results.append(method.result.decl)

        error_var = PythonVar(ERROR_NAME, None,
                              ctx.module.global_module.classes['Exception'])
        error_var.process(ERROR_NAME, self.translator)
        optional_error_var = error_var if method.declared_exceptions else None

        if method.declared_exceptions:
            results.append(error_var.decl)

        mname = ctx.module.get_fresh_name(cls.name + '_' + method.name +
                                          '_inherit_check')
        pres, posts = self.extract_contract(method, ERROR_NAME,
                                            False, ctx)
        if method.method_type == MethodType.normal:
            not_null = self.viper.NeCmp(next(iter(method.args.values())).ref(),
                                        self.viper.NullLit(self.no_position(ctx),
                                                           self.no_info(ctx)),
                                        self.no_position(ctx), self.no_info(ctx))
            new_type = self.type_factory.type_check(
                next(iter(method.args.values())).ref(), cls, pos, ctx,
                concrete=True)
            pres = [not_null, new_type] + pres

        stmts, end_lbl = self.inline_method(method, args, method.result,
                                            error_var, ctx)

        self._create_inherit_check_postamble(stmts, end_lbl, ctx)

        locals_after = set(method.locals.values())
        locals_diff = locals_after.symmetric_difference(locals_before)
        locals = [var.decl for var in locals_diff]
        result = self.create_method_node(
            ctx, mname, params, results, pres, posts, locals, stmts,
            self.no_position(ctx), self.no_info(ctx),
            method=method, overriding_check=True)

        ctx.current_function = old_function

        ctx.position.pop()
        self.info = None
        return result

    def create_override_check(self, method: PythonMethod,
                              ctx: Context) -> 'silver.ast.Callable':
        """
        Creates a Viper function/method with the contract of the overridden
        function which calls the overriding function, to check behavioural
        subtyping.
        """
        assert not method.pure
        old_function = ctx.current_function
        ctx.current_function = method.overrides
        pos = self.viper.to_position(method.node, ctx.position)
        ctx.position.append(('override', pos))
        self.info = self.viper.SimpleInfo(['behavioural.subtyping'])
        self._check_override_validity(method, ctx)

        params = []
        args = []

        for arg in method.overrides.args:
            params.append(method.overrides.args[arg].decl)
            args.append(method.overrides.args[arg].ref())

        self.bind_type_vars(method.overrides, ctx)

        mname = ctx.module.get_fresh_name(method.sil_name + '_override_check')
        pres, posts = self.extract_contract(method.overrides, '_err',
                                            False, ctx)
        self_arg = None
        has_subtype = None
        if method.cls and method.method_type == MethodType.normal:
            self_arg = method.overrides.args[next(iter(method.overrides.args))]
            not_null = self.viper.NeCmp(next(iter(method.overrides.args.values())).ref(),
                                        self.viper.NullLit(
                                            self.no_position(ctx),
                                            self.no_info(ctx)),
                                        self.no_position(ctx),
                                        self.no_info(ctx))
            pres = [not_null] + pres
            has_subtype = self.var_type_check(self_arg.sil_name, method.cls,
                                              pos,
                                              ctx, inhale_exhale=False)
        elif method.method_type == MethodType.class_method:
            cls_arg = next(iter(method.overrides.args.values())).ref()
            has_subtype = self.type_factory.subtype_check(cls_arg, method.cls,
                                                          pos, ctx)
        if method.name == '__init__':
            fields = method.cls.all_fields
            pres.extend([self.get_may_set_predicate(self_arg.ref(), f, ctx)
                         for f in fields])

        called_name = method.sil_name
        ctx.position.pop()
        results, targets, body = self._create_override_check_body_impure(
            method, has_subtype, called_name, args, ctx)
        result = self.create_method_node(
            ctx, mname, params, results, pres, posts, [], body,
            self.no_position(ctx), self.no_info(ctx),
            method=method.overrides, overriding_check=True)

        ctx.current_function = old_function
        self.info = None
        return result

    def _create_override_check_body_impure(self, method: PythonMethod,
            has_subtype: Expr, calledname: str,
            args: List[Expr], ctx: Context) -> Tuple[List['ast.LocalVarDecl'],
                                                     List['ast.LocalVar'],
                                                     List[Stmt]]:
        results = []
        targets = []
        if method.type:
            type = self.translate_type(method.type, ctx)
            result_var_decl = self.viper.LocalVarDecl(RESULT_NAME, type,
                self.to_position(method.node, ctx), self.no_info(ctx))
            result_var_ref = self.viper.LocalVar(RESULT_NAME, type,
                self.to_position(method.node, ctx), self.no_info(ctx))
            results.append(result_var_decl)
            targets.append(result_var_ref)
        error_var_decl = self.viper.LocalVarDecl(ERROR_NAME, self.viper.Ref,
                                                 self.no_position(ctx),
                                                 self.no_info(ctx))
        error_var_ref = self.viper.LocalVar(ERROR_NAME, self.viper.Ref,
                                            self.no_position(ctx),
                                            self.no_info(ctx))
        if method.overrides.declared_exceptions:
            results.append(error_var_decl)
        if method.declared_exceptions:
            targets.append(error_var_ref)

        # Check that arg names match and default args are equal
        default_checks = []
        for (name1, arg1), (name2, arg2) in zip(method.args.items(),
                                                method.overrides.args.items()):
            error_string = ('"default value matches overridden method '
                            'for argument {0}"').format(name1)
            assert_pos = self.to_position(arg1.node, ctx, error_string)
            if name1 != name2:
                raise InvalidProgramException(arg1.node, 'invalid.override')
            if arg1.default or arg2.default:
                if not (arg1.default and arg2.default):
                    raise InvalidProgramException(arg1.node, 'invalid.override')
                val1 = arg1.default_expr
                val2 = arg2.default_expr
                eq = self.viper.EqCmp(val1, val2, assert_pos,
                                      self.no_info(ctx))
                assertion = self.viper.Assert(eq, assert_pos,
                                              self.no_info(ctx))
                default_checks.append(assertion)
        ctx.position.append(('overridden method',
                             self.viper.to_position(method.overrides.node,
                                                    ctx.position)))
        call = self.create_method_call_node(
            ctx, calledname, args, targets,
            self.to_position(method.node, ctx), self.no_info(ctx),
            target_method=method)
        ctx.position.pop()
        if has_subtype:
            subtype_assume = self.viper.Inhale(has_subtype,
                                               self.no_position(ctx),
                                               self.no_info(ctx))
            body = default_checks + [subtype_assume] + call
        else:
            body = default_checks + call
        return results, targets, body

    def _check_override_validity(self, method: PythonMethod,
                                 ctx: Context) -> None:
        """
        Checks if the given method overrides its equivalent in a superclass
        in a valid way, otherwise raises an InvalidProgramException.
        """
        if len(method.args) != len(method.overrides.args):
            raise InvalidProgramException(method.node, 'invalid.override')
        for exc_class in method.declared_exceptions:
            allowed = False
            for superexc in method.overrides.declared_exceptions:
                if exc_class.issubtype(superexc):
                    allowed = True
                    break
            if not allowed:
                raise InvalidProgramException(method.node, 'invalid.override')
                # TODO check if exceptional postconditions imply super postconds
        if method.pure:
            if not method.overrides.pure:
                raise InvalidProgramException(method.node, 'invalid.override')
        else:
            if method.overrides.pure:
                raise InvalidProgramException(method.node, 'invalid.override')

    def translate_default_args(self, method: PythonMethod,
                               ctx: Context) -> None:
        definition_deps = method.definition_deps
        if method.cls:
            definition_deps = method.cls.definition_deps
        for arg in method.args.values():
            if (arg.node and arg.node.annotation and
                    not isinstance(arg.node.annotation, (ast.Str, ast.NameConstant))):
                type = self.get_target(arg.node.annotation, ctx)
                if type and not type.python_class.interface:
                    definition_deps.add((arg.node.annotation, type.python_class,
                                         method.module))
            if arg.default:
                stmt, expr = self.translate_expr(arg.default, ctx)
                if not stmt and expr:
                    arg.default_expr = expr
        if (method.node and method.node.returns and
                not isinstance(method.node.returns, (ast.Str, ast.NameConstant))):
            type = self.get_target(method.node.returns, ctx)
            if type and not type.python_class.interface:
                definition_deps.add((method.node.returns, type.python_class,
                                     method.module))


    def _create_predefined_fields(self,
                                  ctx: Context) -> List[Field]:
        """
        Creates and returns fields needed for encoding various language
        features, e.g. collections, measures and iterators.
        """
        fields = []
        fields.append(self.viper.Field(GLOBAL_VAR_FIELD, self.viper.Ref,
                                       self.no_position(ctx),
                                       self.no_info(ctx)))
        fields.append(self.viper.Field('__container', self.viper.Ref,
                                       self.no_position(ctx),
                                       self.no_info(ctx)))
        fields.append(self.viper.Field('__iter_index', self.viper.Int,
                                       self.no_position(ctx),
                                       self.no_info(ctx)))
        fields.append(self.viper.Field('__previous', self.viper.SeqType(self.viper.Ref),
                                       self.no_position(ctx),
                                       self.no_info(ctx)))
        fields.append(self.viper.Field('list_acc',
                                       self.viper.SeqType(self.viper.Ref),
                                       self.no_position(ctx),
                                       self.no_info(ctx)))
        fields.append(self.viper.Field('set_acc',
                                       self.viper.SetType(self.viper.Ref),
                                       self.no_position(ctx),
                                       self.no_info(ctx)))
        fields.append(self.viper.Field('dict_acc',
                                       self.viper.SetType(self.viper.Ref),
                                       self.no_position(ctx),
                                       self.no_info(ctx)))
        fields.append(self.viper.Field('Measure$acc',
                                       self.viper.SeqType(self.viper.Ref),
                                       self.no_position(ctx),
                                       self.no_info(ctx)))
        return fields

    def _add_all_used_names(self, initial: Set[str]) -> None:
        """
        Calculates the names of all methods and functions used by the program,
        based on the names reported to be used by the viper_ast module, and adds
        them to the given set.
        """
        used_names = initial
        to_add = list(self.viper.used_names)
        index = 0
        while index < len(to_add):
            current = to_add[index]
            if current not in used_names:
                used_names.add(current)
                if current in self.required_names:
                    to_add.extend(self.required_names[current])
            index = index + 1

    def _convert_silver_elements(
            self, sil_progs: Program, all_used: List[str],
            ctx: Context) -> Tuple[List[Domain],
                                   List[Predicate],
                                   List[Function],
                                   List[Method]]:
        """
        Extracts domains, functions, predicates and methods from the given list
        of Silver programs, applies the necessary conversions (e.g. related to
        obligations) to them, and returns them in separate lists.
        """
        domains = []
        functions = []
        predicates = []
        methods = []

        if all_used:
            used_names = set(all_used)
            self.viper.used_names = set()
        else:
            used_names = set()
            self._add_all_used_names(used_names)

        # Reset used names set, we only need the additional ones used by the
        # upcoming method transformation.
        self.viper.used_names = set()
        for method in self.viper.to_list(sil_progs.methods()):
            if method.name() in used_names:
                body = self.viper.from_option(method.body())
                converted_method = self.create_method_node(
                    ctx=ctx,
                    name=method.name(),
                    args=self.viper.to_list(method.formalArgs()),
                    returns=self.viper.to_list(method.formalReturns()),
                    pres=self.viper.to_list(method.pres()),
                    posts=self.viper.to_list(method.posts()),
                    locals=[],
                    body=body,
                    position=method.pos(),
                    info=method.info(),
                )
                methods.append(converted_method)

        # Some obligation-related functions may only be used by the code added
        # by the method conversion we just performed, so we have to add
        # the names which have been used in the meantime. This works assuming
        # that the converted code does not introduce additional method
        # requirements (which should never be the case).
        self._add_all_used_names(used_names)

        domains += [
            domain for domain in self.viper.to_list(sil_progs.domains())
            if domain.name() != 'PyType']

        functions += [
            function
            for function in self.viper.to_list(sil_progs.functions())
            if function.name() in used_names]
        predicates += self.viper.to_list(sil_progs.predicates())

        return domains, predicates, functions, methods

    def track_dependencies(self, selected_names: List[str], selected: Set[str],
                           node: PythonNode, ctx: Context) -> None:
        """
        If specific parts of the program have been selected to be verified,
        marks that the given PythonNode is about to be translated, s.t. it can
        be tracked which other elements are referenced by the translation of
        this node. Also checks if the given element is among those selected
        to be verified, and adds its Silver name to the list of selected Silver
        names later used when computing which parts of the program to give to
        Viper.
        """
        if not selected:
            return
        if node.sil_name in self.viper.used_names_sets:
            used_names = self.viper.used_names_sets[node.sil_name]
        else:
            used_names = set()
        self.viper.used_names = used_names
        self.viper.used_names_sets[node.sil_name] = used_names
        if selected_names is None:
            return
        if (node.name in selected or
                (hasattr(node, 'cls') and node.cls and
                 node.cls.name + '.' + node.name in selected)):
            selected_names.append(node.sil_name)

    def create_functions_domain(self, constants: List, ctx: Context):
        return self.viper.Domain(FUNCTION_DOMAIN_NAME, constants, [], [],
                                 self.no_position(ctx), self.no_info(ctx))

    def translate_function_constant(self, func: PythonMethod, ctx: Context):
        func_type = self.viper.function_domain_type()
        return self.viper.DomainFunc(func.func_constant, [], func_type, True,
                                     self.to_position(func.node, ctx), self.no_info(ctx),
                                     FUNCTION_DOMAIN_NAME)

    def create_joinable_function(self, ctx: Context) -> Function:
        thread_arg_decl = self.viper.LocalVarDecl('t', self.viper.Ref,
                                                  self.no_position(ctx),
                                                  self.no_info(ctx))
        return self.viper.Function(JOINABLE_FUNC, [thread_arg_decl], self.viper.Bool,
                                   [], [], None, self.no_position(ctx), self.no_info(ctx))

    def create_thread_predicates(self, ctx: Context) -> Function:
        thread_arg_decl = self.viper.LocalVarDecl('t', self.viper.Ref,
                                                  self.no_position(ctx),
                                                  self.no_info(ctx))
        post_pred =  self.viper.Predicate(THREAD_POST_PRED, [thread_arg_decl], None,
                                          self.no_position(ctx), self.no_info(ctx))
        start_pred = self.viper.Predicate(THREAD_START_PRED, [thread_arg_decl], None,
                                          self.no_position(ctx), self.no_info(ctx))
        return [start_pred, post_pred]

    def create_method_id_domain(self, constants: List['silver.ast.DomainFunc'],
                                ctx: Context) -> 'silver.ast.Domain':
        return self.viper.Domain(METHOD_ID_DOMAIN, constants, [], [],
                                 self.no_position(ctx), self.no_info(ctx))

    def translate_method_id_to_constant(self, method, ctx) -> 'silver.ast.DomainFunc':
        func_type = self.viper.DomainType(METHOD_ID_DOMAIN, {}, [])
        return self.viper.DomainFunc(method.threading_id,[],func_type, True,
                                     self.to_position(method.node,ctx), self.no_info(ctx),
                                     METHOD_ID_DOMAIN)

    def create_thread_domain(self, ctx: Context) -> 'silver.ast.Domain':
        pos, info = self.no_position(ctx), self.no_info(ctx)
        method_id_type = self.viper.DomainType(METHOD_ID_DOMAIN, {}, [])
        thread_param = self.viper.LocalVarDecl('t', self.viper.Ref, pos, info)
        index_param = self.viper.LocalVarDecl('i', self.viper.Int, pos, info)
        get_method = self.viper.DomainFunc(GET_METHOD_FUNC, [thread_param],
                                           method_id_type, False, pos, info,
                                           THREAD_DOMAIN)
        get_arg = self.viper.DomainFunc(GET_ARG_FUNC, [thread_param, index_param],
                                        self.viper.Ref, False, pos, info, THREAD_DOMAIN)
        get_old = self.viper.DomainFunc(GET_OLD_FUNC, [thread_param, index_param],
                                       self.viper.Ref, False, pos, info, THREAD_DOMAIN)
        domain = self.viper.Domain(THREAD_DOMAIN, [get_method, get_arg, get_old], [], [],
                                   pos, info)
        if isinstance(self.viper, ViperASTExtended):
            self.viper.ast_extensions.SIFExtendedTransformer.addDomainFuncToDuplicate(
                self.viper.to_seq([get_method, get_arg, get_old]))
        return domain

    def create_definedness_functions(self, ctx: Context) -> List['silver.ast.Function']:
        pos = self.no_position(ctx)
        info = self.no_info(ctx)
        id_param_decl = self.viper.LocalVarDecl('id', self.viper.Int, pos, info)
        id_param = self.viper.LocalVar('id', self.viper.Int, pos, info)
        is_defined_func = self.viper.Function(IS_DEFINED_FUNC, [id_param_decl],
                                              self.viper.Bool, [], [], None, pos, info)
        var_param_decl = self.viper.LocalVarDecl('val', self.viper.Ref, pos, info)
        var_param = self.viper.LocalVar('val', self.viper.Ref, pos, info)
        is_defined_pre = self.viper.FuncApp(IS_DEFINED_FUNC, [id_param], pos, info,
                                            self.viper.Bool, [id_param_decl])
        check_defined_func = self.viper.Function(CHECK_DEFINED_FUNC,
                                                 [var_param_decl, id_param_decl],
                                                 self.viper.Ref, [is_defined_pre], [],
                                                 var_param, pos, info)
        return [is_defined_func, check_defined_func]

    def create_asserting_function(self,
                                            ctx: Context) -> List['silver.ast.Function']:
        pos = self.no_position(ctx)
        info = self.no_info(ctx)
        var_param_decl = self.viper.LocalVarDecl('val', self.viper.Ref, pos, info)
        var_param = self.viper.LocalVar('val', self.viper.Ref, pos, info)
        assertion_param_decl = self.viper.LocalVarDecl('ass', self.viper.Bool, pos, info)
        assertion_param = self.viper.LocalVar('ass', self.viper.Bool, pos, info)
        asserting_func = self.viper.Function(ASSERTING_FUNC,
                                             [var_param_decl, assertion_param_decl],
                                             self.viper.Ref, [assertion_param], [],
                                             var_param, pos, info)
        return [asserting_func]

    def create_arbitrary_bool_func(self, ctx: Context) -> 'silver.ast.Function':
        pos = self.no_position(ctx)
        info = self.no_info(ctx)
        i_param_decl = self.viper.LocalVarDecl('i', self.viper.Int, pos, info)
        return self.viper.Function(ARBITRARY_BOOL_FUNC, [i_param_decl],
                                   self.viper.Bool, [], [], None, pos, info)

    def create_may_set_predicate(self, ctx: Context) -> 'silver.ast.Predicate':
        pos = self.no_position(ctx)
        info = self.no_info(ctx)
        receiver_param_decl = self.viper.LocalVarDecl('rec', self.viper.Ref, pos, info)
        id_param_decl = self.viper.LocalVarDecl('id', self.viper.Int, pos, info)
        may_set_pred = self.viper.Predicate(MAY_SET_PRED,
                                            [receiver_param_decl, id_param_decl], None,
                                            pos, info)
        return may_set_pred

    def _get_adt_cons_params_decl(self, cons: PythonClass, adt_name: str,
                                  adt_type: DomainType, pos: Position,
                                 info: Info, ctx: Context) -> List[VarDecl]:
        """
        Returns the parameters of the ADT constructor as a list of variable
        declarations.
        """
        arguments = []
        for arg_name, arg_type in cons.fields.items():
            if arg_type.type.name == adt_name:
                argument_type = adt_type
            else:
                argument_type = self.translate_type(arg_type.type, ctx)
            argument = self.viper.LocalVarDecl(arg_name, argument_type, pos, info)
            arguments.append(argument)
        return arguments

    def _get_adt_cons_params_use(self, cons: PythonClass, adt_name: str,
                                 adt_type: DomainType, pos: Position,
                                 info: Info, ctx: Context) -> List[Var]:
        """
        Returns the parameters of the ADT constructor as a list of variables.
        """
        arguments = []
        for arg_name, arg_type in cons.fields.items():
            if arg_type.type.name == adt_name:
                argument_type = adt_type
            else:
                argument_type = self.translate_type(arg_type.type, ctx)
            argument = self.viper.LocalVar(arg_name, argument_type, pos, info)
            arguments.append(argument)
        return arguments

    def _conjoin(self, eqs: List[Expr], pos: Position, info: Info) -> Expr:
        """
        Conjoin all expressions in the list.
        """
        return eqs[0] if len(eqs) == 1 else self.viper.And(eqs[0],
               self._conjoin(eqs[1:], pos, info), pos, info)

    def _disjoin(self, eqs: List[Expr], pos: Position, info: Info) -> Expr:
        """
        Disjoin all expressions in the list.
        """
        return eqs[0] if len(eqs) == 1 else self.viper.Or(eqs[0],
               self._disjoin(eqs[1:], pos, info), pos, info)

    def _create_adt_func_constructors(self, adt: PythonClass, adt_type: DomainType,
                                      pos: Position, info: Info,
                                      ctx: Context) -> List[DomainFunc]:
        """
        Create domain functions representing constructors of the ADT.
        """
        assert adt.all_subclasses[0] == adt
        for cons in adt.all_subclasses[1:]:
            arguments = self._get_adt_cons_params_decl(cons, adt.name, adt_type, pos,
                                                       info, ctx)
            yield self.viper.DomainFunc(adt.fresh(adt.adt_prefix + cons.name),
                                        arguments, adt_type, False, pos, info,
                                        adt.adt_domain_name)
    
    def _create_adt_func_deconstructors(self, adt: PythonClass, adt_type: DomainType,
                                        pos: Position, info: Info,
                                        ctx: Context) -> List[DomainFunc]:
        """
        Create domain functions representing deconstructors of the ADT.
        """
        adt_obj_decl = self.viper.LocalVarDecl('obj', adt_type, pos, info)
        for cons in adt.all_subclasses[1:]:
            for arg_name, arg_type in cons.fields.items():
                if arg_type.type.name == adt.name:
                    function_type = adt_type
                else:
                    function_type = self.translate_type(arg_type.type, ctx)
                yield self.viper.DomainFunc(adt.fresh(adt.adt_prefix + cons.name
                                            + '_' + arg_name), [adt_obj_decl],
                                            function_type, False, pos, info,
                                            adt.adt_domain_name)

    def _create_adt_func_constructor_types(self, adt: PythonClass, adt_type: DomainType,
                                           pos: Position, info: Info) -> List[DomainFunc]:
        """
        Create domain function representing the constructor types of the ADT.
        """
        # Given the ADT, return the constructor used to create it
        adt_obj_decl = self.viper.LocalVarDecl('obj', adt_type, pos, info)
        yield self.viper.DomainFunc(adt.fresh(adt.adt_prefix + 'cons_type'),
                                    [adt_obj_decl], self.viper.Int, False,
                                    pos, info, adt.adt_domain_name)

        # Create a constant for each constructor
        for cons in adt.all_subclasses[1:]:
            yield self.viper.DomainFunc(adt.fresh(adt.adt_prefix + cons.name +
                                        '_type'), [], self.viper.Int, True, pos,
                                        info, adt.adt_domain_name)

        # Create a boolean function for each type
        for cons in adt.all_subclasses[1:]:
            yield self.viper.DomainFunc(adt.fresh(adt.adt_prefix + 'is_' +
                                        cons.name), [adt_obj_decl],
                                        self.viper.Bool, False, pos, info,
                                        adt.adt_domain_name)

    def _create_adt_axiom_deconstructors_over_constructors(self, adt: PythonClass,
                                      adt_type: DomainType, pos: Position,
                                      info: Info, ctx: Context) -> List[DomainAxiom]:
        """
        Create domain axiom representing the distribution of deconstructors over
        constructors.
        """
        for cons in adt.all_subclasses[1:]:
            args = self._get_adt_cons_params_use(cons, adt.name, adt_type, pos, info, ctx)
            args_decl = self._get_adt_cons_params_decl(cons, adt.name, adt_type, pos,
                                                       info, ctx)
            if len(args) > 0:
                cons_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix +
                                                     cons.name), args, adt_type,
                                                     pos, info, adt.adt_domain_name)
                triggers = []
                eqs = []
                for (arg_name, _), arg in zip(cons.fields.items(), args):
                    decons_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix +
                                                           cons.name + '_' + arg_name),
                                                           [cons_call], arg.typ(), pos,
                                                           info, adt.adt_domain_name)
                    eq = self.viper.EqCmp(decons_call, arg, pos, info)
                    eqs.append(eq)
                    trigger = self.viper.Trigger([decons_call], pos, info)
                    triggers.append(trigger)
                body = self._conjoin(eqs, pos, info)
                forall = self.viper.Forall(args_decl, triggers, body, pos, info)
                yield self.viper.DomainAxiom(adt.fresh(adt.adt_prefix +
                                             'decons_over_cons_' + cons.name),
                                             forall, pos, info, adt.adt_domain_name)

    def _create_adt_axiom_constructors_over_deconstructors(self, adt: PythonClass,
                                      adt_type: DomainType, pos: Position,
                                      info: Info, ctx: Context) -> List[DomainAxiom]:
        """
        Create domain axiom representing the distribution of constructors over
        deconstructors.
        """
        adt_obj_use = self.viper.LocalVar('obj', adt_type, pos, info)
        adt_obj_decl = self.viper.LocalVarDecl('obj', adt_type, pos, info)
        for cons in adt.all_subclasses[1:]:
            if len(cons.fields.items()) > 0:
                args = self._get_adt_cons_params_use(cons, adt.name, adt_type, pos, info,
                                                     ctx)
                triggers, decons_calls = [], []
                is_cons_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix
                                                        + 'is_' + cons.name),
                                                        [adt_obj_use],
                                                        self.viper.Bool,
                                                        pos, info,
                                                        adt.adt_domain_name)
                for (arg_name, _), arg in zip(cons.fields.items(), args):
                    decons_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix
                                                           + cons.name + '_' + arg_name),
                                                           [adt_obj_use], arg.typ(), pos,
                                                           info, adt.adt_domain_name)
                    decons_calls.append(decons_call)
                    triggers.append(self.viper.Trigger([decons_call], pos, info))
                cons_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix + cons.name),
                                                     decons_calls, adt_type, pos, info,
                                                     adt.adt_domain_name)
                eq = self.viper.EqCmp(adt_obj_use, cons_call, pos, info)
                body = self.viper.Implies(is_cons_call, eq, pos, info)
                adt_obj_decl = self.viper.LocalVarDecl('obj', adt_type, pos, info)
                trigger = self.viper.Trigger([is_cons_call], pos, info)
                forall = self.viper.Forall([adt_obj_decl], [trigger], body, pos, info)
                yield self.viper.DomainAxiom(adt.fresh(adt.adt_prefix + 'cons_'
                                             + cons.name + '_over_decons'), forall,
                                             pos, info, adt.adt_domain_name)

    def _create_adt_axiom_associate_cons_type_with_const(self, adt: PythonClass,
                                      adt_type: DomainType, pos: Position,
                                      info: Info, ctx: Context) -> List[DomainAxiom]:
        """
        Create domain axiom associating each construtor type with a
        respective constant.
        """
        adt_obj_use = self.viper.LocalVar('obj', adt_type, pos, info)
        for cons in adt.all_subclasses[1:]:
            args = self._get_adt_cons_params_use(cons, adt.name, adt_type, pos, info, ctx)
            cons_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix +
                                                 cons.name), args, adt_type, pos,
                                                 info, adt.adt_domain_name)
            cons_type_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix +
                                                      'cons_type'), [cons_call],
                                                      self.viper.Int, pos, info,
                                                      adt.adt_domain_name)
            cons_const_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix
                                                       + cons.name + '_type'), [],
                                                       self.viper.Int, pos, info,
                                                       adt.adt_domain_name)
            eq = self.viper.EqCmp(cons_type_call, cons_const_call, pos, info)
            if len(cons.fields.items()) == 0:
                forall = eq
            else:
                args_decl = self._get_adt_cons_params_decl(cons, adt.name, adt_type, pos,
                                                           info, ctx)
                trigger = self.viper.Trigger([cons_call], pos, info)
                forall = self.viper.Forall(args_decl, [trigger], eq, pos, info)
            yield self.viper.DomainAxiom(adt.fresh(adt.adt_prefix +
                                         'associate_cons_type_function_with_' +
                                         cons.name + '_constant'), forall, pos,
                                         info, adt.adt_domain_name)

    def _create_adt_axiom_constrain_cons_type_with_const(self, adt: PythonClass,
                                                         adt_type: DomainType,
                                                         pos: Position,
                                                         info: Info
                                                         ) -> List[DomainAxiom]:
        """
        Create domain axiom constraining each construtor type to a
        respective constant.
        """
        adt_obj_use = self.viper.LocalVar('obj', adt_type, pos, info)
        adt_obj_decl = self.viper.LocalVarDecl('obj', adt_type, pos, info)
        eqs = []
        for cons in adt.all_subclasses[1:]:
            cons_type_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix
                                                      + 'cons_type'), [adt_obj_use],
                                                        self.viper.Int, pos, info,
                                                        adt.adt_domain_name)
            cons_const_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix +
                                                       cons.name + '_type'), [],
                                                        self.viper.Int, pos, info,
                                                        adt.adt_domain_name)
            eqs.append(self.viper.EqCmp(cons_type_call, cons_const_call, pos, info))
        body = self._disjoin(eqs, pos, info)
        trigger = self.viper.Trigger([cons_type_call], pos, info)
        forall = self.viper.Forall([adt_obj_decl], [trigger], body, pos, info)
        yield self.viper.DomainAxiom(adt.fresh(adt.adt_prefix +
                                     'constrain_cons_type_function_cons_constants'),
                                     forall, pos, info, adt.adt_domain_name)

    def _create_adt_axiom_associate_cons_type_with_bool(self, adt: PythonClass,
                                                        adt_type: DomainType,
                                                        pos: Position,
                                                        info: Info
                                                        ) -> List[DomainAxiom]:
        """
        Create domain axiom associating each construtor type to a
        boolean function that returns true when the ADT was created
        by the constructor the boolean function refers to.
        """
        adt_obj_use = self.viper.LocalVar('obj', adt_type, pos, info)
        adt_obj_decl = self.viper.LocalVarDecl('obj', adt_type, pos, info)
        for cons in adt.all_subclasses[1:]:
            cons_type_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix +
                                                      'cons_type'), [adt_obj_use],
                                                      self.viper.Int, pos, info,
                                                      adt.adt_domain_name)
            cons_const_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix +
                                                       cons.name + '_type'), [],
                                                       self.viper.Int, pos, info,
                                                       adt.adt_domain_name)
            type_id_eq = self.viper.EqCmp(cons_type_call, cons_const_call, pos, info)
            is_cons_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix + 'is_'
                                                    + cons.name), [adt_obj_use],
                                                    self.viper.Bool, pos, info,
                                                    adt.adt_domain_name)
            eqv = self.viper.EqCmp(type_id_eq, is_cons_call, pos, info)
            trigger = self.viper.Trigger([cons_type_call], pos, info)
            forall = self.viper.Forall([adt_obj_decl], [trigger], eqv, pos, info)
            yield self.viper.DomainAxiom(adt.fresh(adt.adt_prefix +
                                         'associate_cons_type_function_with_is_'
                                         + cons.name + '_bool_function'), forall,
                                         pos, info, adt.adt_domain_name)

    def _create_adt_axiom_cons_are_subclasses(self, adt: PythonClass,
                                              adt_type: DomainType, pos: Position,
                                              info: Info, ctx: Context
                                              ) -> List[DomainAxiom]:
        """
        Create domain axiom enforcing the fact that each constructor is a subclass
        of the class defining the ADT type. This axiom applies to PyType domain
        specifically.
        """
        ref_var_use = self.viper.LocalVar('ref', self.viper.Ref, pos, info)
        typeof_expr = self.type_factory.typeof(ref_var_use, ctx)
        func_call_expr = self.viper.DomainFuncApp(adt.name, [],
                                                  self.type_factory.type_type(),
                                                  pos, info,
                                                  self.type_factory.type_domain)
        issubtype_expr = self.type_factory._issubtype(typeof_expr, func_call_expr,
                                                      ctx)
        eqs = []
        for cons in adt.all_subclasses[1:]:
            cons_call = self.viper.DomainFuncApp(cons.name, [],
                                                 self.type_factory.type_type(),
                                                 pos, info,
                                                 self.type_factory.type_domain)
            eq = self.viper.EqCmp(typeof_expr, cons_call, pos, info)
            eqs.append(eq)
        or_expr = self._disjoin(eqs, pos, info)
        impl = self.viper.Implies(issubtype_expr, or_expr, pos, info)
        ref_var_decl = self.viper.LocalVarDecl('ref', self.viper.Ref, pos, info)
        trigger = self.viper.Trigger([issubtype_expr], pos, info)
        forall = self.viper.Forall([ref_var_decl], [trigger], impl, pos, info)
        yield self.viper.DomainAxiom(adt.fresh(adt.adt_prefix +
                                     'type_of_constructors'), forall, pos, info,
                                     adt.adt_domain_name)

    def _create_adt_func_box_and_unbox(self, adt: PythonClass, adt_type: DomainType,
                                       pos: Position, info: Info, ctx: Context
                                       ) -> List[Function]:
        """
        Create functions representing boxing and unboxing of the ADT to
        and from Ref, respectively.
        """
        ## Create box function
        adt_obj_use = self.viper.LocalVar('obj', adt_type, pos, info)
        adt_obj_decl = self.viper.LocalVarDecl('obj', adt_type, pos, info)
        postconds = []
        result = self.viper.Result(self.viper.Ref, pos, info)
        postconds.append(self.type_factory.type_check(result, adt, pos, ctx))
        unbox_func = self.viper.FuncApp(adt.fresh('unbox_' + adt.adt_domain_name),
                                        [result], pos, info, adt_type)
        postconds.append(self.viper.EqCmp(unbox_func, adt_obj_use, pos, info))
        for cons in adt.all_subclasses[1:]:
            is_cons_call = self.viper.DomainFuncApp(adt.fresh(adt.adt_prefix + 'is_'
                                                    + cons.name), [adt_obj_use],
                                                    self.viper.Bool, pos, info,
                                                    adt.adt_domain_name)
            typeof_call = self.type_factory.typeof(result, ctx)
            const_call = self.viper.DomainFuncApp(cons.name, [],
                         self.type_factory.type_type(), pos, info,
                         self.type_factory.type_domain)
            typeof_eq = self.viper.EqCmp(typeof_call, const_call, pos, info)
            postconds.append(self.viper.Implies(is_cons_call, typeof_eq, pos, info))
        yield self.viper.Function(adt.fresh('box_' + adt.adt_domain_name),
                                  [adt_obj_decl], self.viper.Ref, [], postconds,
                                  None, pos, info)

        ## Create unbox function
        preconds = []
        postconds = []
        adt_ref_use = self.viper.LocalVar('ref', self.viper.Ref, pos, info)
        preconds.append(self.type_factory.type_check(adt_ref_use, adt, pos, ctx))
        result = self.viper.Result(adt_type, pos, info)
        box_func = self.viper.FuncApp(adt.fresh('box_' + adt.adt_domain_name),
                                      [result], pos, info, self.viper.Ref)
        postconds.append(self.viper.EqCmp(box_func, adt_ref_use, pos, info))
        adt_ref_decl = self.viper.LocalVarDecl('ref', self.viper.Ref, pos, info)
        yield self.viper.Function(adt.fresh('unbox_' + adt.adt_domain_name),
                                  [adt_ref_decl], adt_type, preconds, postconds,
                                  None, pos, info)

    def create_adts_domains_and_functions(self, adts: List[PythonClass],
                                          ctx: Context) -> List['silver.ast.domain']:
        """
        Translate Algebraic Data Types defined in Python, with classes (sum)
        and a NamedTuple (product), to a domain in Viper. Further documentation
        on ADT syntax can be found in ADT.py.
        """

        info = self.no_info(ctx)
        domains = []
        functions = []

        for adt in adts:
            assert adt.is_adt and adt.is_defining_adt

            # ADTs should have constructors
            if not len(adt.all_subclasses) > 1:
                raise InvalidProgramException(adt.node, 'malformed.adt',
                    'malformed algebraic datatype: ADT has no constructors, ' +
                    'which should be defined as subclasses of the ADT class')

            # Create a domain type for each ADT and its constructors and
            # deconstructors
            adt_type = self.viper.DomainType(adt.adt_domain_name, {}, [])

            # All positions refer to ADT definition
            pos = self.to_position(adt, ctx)

            # Create domain functions
            domain_funcs = []

            ## Create constructors
            domain_funcs.extend(self._create_adt_func_constructors(adt,
                                adt_type, pos, info, ctx))

            ## Create deconstructors
            domain_funcs.extend(self._create_adt_func_deconstructors(adt,
                                adt_type, pos, info, ctx))
            
            ## Constructor types
            domain_funcs.extend(self._create_adt_func_constructor_types(adt,
                                adt_type, pos, info))

            # Create domain axioms
            axioms = []

            ## Destructors over constructors
            axioms.extend(self._create_adt_axiom_deconstructors_over_constructors(
                          adt, adt_type, pos, info, ctx))

            ## Constructors over destructors
            axioms.extend(self._create_adt_axiom_constructors_over_deconstructors(
                          adt, adt_type, pos, info, ctx))

            ## Associate constructor type function with constructor constant
            axioms.extend(self._create_adt_axiom_associate_cons_type_with_const(
                          adt, adt_type, pos, info, ctx))

            ## Constrain constructor type function to constructor constants
            axioms.extend(self._create_adt_axiom_constrain_cons_type_with_const(
                          adt, adt_type, pos, info))

            ## Associate constructor type function with constructor boolean function
            axioms.extend(self._create_adt_axiom_associate_cons_type_with_bool(
                          adt, adt_type, pos, info))

            ## Express the fact that constructors can only be subclasses of the
            ## ADT (abstract class)
            axioms.extend(self._create_adt_axiom_cons_are_subclasses(
                          adt, adt_type, pos, info, ctx))

            # Create domain
            domains.append(self.viper.Domain(adt.adt_domain_name, domain_funcs,
                           axioms, [], pos, info))

            # Create ADT boxing and unboxing functions
            functions.extend(self._create_adt_func_box_and_unbox(adt, adt_type,
                             pos, info, ctx))

        return domains, functions

    def translate_program(self, modules: List[PythonModule], sil_progs: Program,
                          ctx: Context, selected: Set[str] = None,
                          ignore_global: bool = False) -> Program:
        """
        Translates the PythonModules created by the analyzer to a Viper program.
        """
        fields = self._create_predefined_fields(ctx)
        domains = []
        predicates = []
        functions = []
        methods = []

        # Predefined obligation stuff
        obl_predicates, obl_fields = self.get_obligation_preamble(ctx)
        predicates.extend(obl_predicates)
        predicates.extend(self.create_thread_predicates(ctx))
        functions.append(self.create_joinable_function(ctx))
        fields.extend(obl_fields)

        functions.extend(self.create_definedness_functions(ctx))
        functions.extend(self.create_asserting_function(ctx))
        functions.append(self.create_arbitrary_bool_func(ctx))
        predicates.append(self.create_may_set_predicate(ctx))

        type_funcs = self.type_factory.get_default_functions(ctx)
        type_axioms = self.type_factory.get_default_axioms(ctx)

        predicate_families = OrderedDict()
        static_fields = OrderedDict()
        func_constants = []

        threading_ids_constants = []
        # Silver names of the set of nodes which have been selected by the user
        # to be verified (if any).
        selected_names = []

        # List of classes that define algebraic data types
        adt_list = []

        # First iteration over all modules: translate global variables, static
        # fields, and default arguments.
        for module in modules:
            ctx.module = module
            containers = [module]
            for class_name, cls in module.classes.items():
                if class_name in PRIMITIVES or class_name != cls.name:
                    # Skip primitives or entries for type variables.
                    continue
                if cls.is_adt:
                    # Prevents fields from being generated
                    continue
                containers.append(cls)
                f_fields, f_funcs, f_methods = self._translate_fields(cls, ctx)
                fields += f_fields
                methods += f_methods
                functions += f_funcs
                ctx.current_class = cls
                for field_name in cls.all_static_fields:
                    field = cls.get_static_field(field_name)
                    current_field = field
                    while current_field.overrides:
                        current_field = current_field.overrides
                    static_fields.setdefault(current_field, []).append(cls)

            ctx.current_class = None
            # Translate default args
            for container in containers:
                for function in container.functions.values():
                    self.track_dependencies(None, selected, function, ctx)
                    self.translate_default_args(function, ctx)
                for method in container.methods.values():
                    self.track_dependencies(None, selected, method, ctx)
                    self.translate_default_args(method, ctx)
                for pred in container.predicates.values():
                    self.track_dependencies(None, selected, pred, ctx)
                    self.translate_default_args(pred, ctx)

        for root, classes in static_fields.items():
            self.track_dependencies(None, selected, root, ctx)
            functions.append(self.create_static_field_function(root, classes,
                                                               ctx))

        # Second iteration over all modules: Translate everything else.
        for module in modules:
            ctx.module = module

            for function in module.functions.values():
                if function.interface:
                    continue
                self.track_dependencies(selected_names, selected, function, ctx)
                functions.append(self.translate_function(function, ctx))
                func_constants.append(self.translate_function_constant(function, ctx))
            for method in module.methods.values():
                id_constant = self.translate_method_id_to_constant(method, ctx)
                threading_ids_constants.append(id_constant)
                if method.interface:
                    continue
                self.track_dependencies(selected_names, selected, method, ctx)
                methods.append(self.translate_method(method, ctx))
            for pred in module.predicates.values():
                self.track_dependencies(selected_names, selected, pred, ctx)
                predicates.append(self.translate_predicate(pred, ctx))
            for class_name, cls in module.classes.items():
                if class_name in PRIMITIVES or class_name != cls.name:
                    # Skip primitives and type variable entries.
                    continue
                if cls.is_adt and cls.is_defining_adt:
                    adt_list.append(cls)
                old_class = ctx.current_class
                ctx.current_class = cls
                funcs, axioms = self.type_factory.create_type(cls, ctx)
                type_funcs.extend(funcs)
                if axioms:
                    type_axioms.extend(axioms)
                for func_name in cls.functions:
                    func = cls.functions[func_name]
                    if func.interface:
                        continue
                    self.track_dependencies(selected_names, selected, func, ctx)
                    functions.append(self.translate_function(func, ctx))
                    func_constants.append(self.translate_function_constant(func, ctx))
                    if func.overrides and not (func_name in ('__str__', '__bool__') and
                                               func.overrides.cls.name == 'object'):
                        # We allow overriding certain methods, since the basic versions
                        # in object are already minimal.
                        raise InvalidProgramException(func.node,
                                                      'invalid.override')
                for method_name in cls.methods:
                    method = cls.methods[method_name]
                    threading_ids_constants.append(
                        self.translate_method_id_to_constant(method, ctx))
                    if method.interface:
                        continue
                    self.track_dependencies(selected_names, selected, method, ctx)
                    methods.append(self.translate_method(method, ctx))
                    if ((method_name != '__init__' or
                             (cls.superclass and
                              cls.superclass.python_class.has_classmethod)) and
                            method.overrides):
                        methods.append(self.create_override_check(method, ctx))
                for method_name in cls.static_methods:
                    method = cls.static_methods[method_name]
                    threading_ids_constants.append(
                        self.translate_method_id_to_constant(method, ctx))
                    self.track_dependencies(selected_names, selected, method, ctx)
                    methods.append(self.translate_method(method, ctx))
                    if method.overrides:
                        methods.append(self.create_override_check(method, ctx))
                for method_name in cls.all_methods:
                    method = cls.get_method(method_name)
                    if (method.cls and method.cls != cls and
                            method_name != '__init__' and
                            method.method_type == MethodType.normal and
                            not method.interface and
                            not method.contract_only):
                        # Inherited
                        methods.append(self.create_inherit_check(method, cls,
                                                                 ctx))
                for pred_name in cls.predicates:
                    pred = cls.predicates[pred_name]
                    cpred = pred
                    while cpred.overrides:
                        cpred = cpred.overrides
                    if cpred in predicate_families:
                        predicate_families[cpred].append(pred)
                    else:
                        predicate_families[cpred] = [pred]
                ctx.current_class = old_class

        if not ignore_global:
            main_py_method, main_method = self.translate_main_method(modules, ctx)
            methods.append(main_method)
            self.track_dependencies(selected_names, selected, main_py_method, ctx)

        # Translate global variables.
        for module in modules:
            ctx.module = module
            for var in module.global_vars.values():
                if not var.module is module:
                    continue
                self.track_dependencies(selected_names, selected, var, ctx)
                # TODO: Check for every references function:
                # - if it doesnt exist yet it has no precondition so its okay
                # - if its precondition contains predicates or perms
                functions.append(
                    self.create_global_var_function(var, ctx))

        # IO operations are translated last because we need to know which functions are
        # used with Eval.
        for module in modules:
            for operation in module.io_operations.values():
                self.track_dependencies(selected_names, selected, operation, ctx)
                predicate, getters, checkers = self.translate_io_operation(
                    operation,
                    ctx)
                predicates.append(predicate)
                functions.extend(getters)
                methods.extend(checkers)

        for root in predicate_families:
            self.track_dependencies(selected_names, selected, root, ctx)
            pf = self.translate_predicate_family(root, predicate_families[root],
                                                 ctx)
            predicates.append(pf)

        all_used_names = None
        if selected:
            # Compute all dependencies of directly selected methods/...
            all_used_names = list(selected_names)
            i = 0
            while i < len(all_used_names):
                name = all_used_names[i]
                to_add = set()
                if name in self.viper.used_names_sets:
                    to_add = self.viper.used_names_sets[name]
                if name in self.required_names:
                    to_add = self.required_names[name]
                for add in to_add:
                    if not add in all_used_names:
                        all_used_names.append(add)
                i += 1

            # Filter out anything the selected part does not depend on.
            predicates = [p for p in predicates if p.name() in all_used_names]
            functions = [f for f in functions if f.name() in all_used_names]
            methods = [m for m in methods if m.name() in all_used_names]

        domains.append(self.type_factory.create_type_domain(type_funcs,
                                                            type_axioms, ctx))
        domains.append(self.create_thread_domain(ctx))
        domains.append(self.create_functions_domain(func_constants, ctx))
        domains.append(self.create_method_id_domain(threading_ids_constants, ctx))
        adts_domains, adts_functions = self.create_adts_domains_and_functions(adt_list,
                                                                              ctx)
        domains.extend(adts_domains)
        functions.extend(adts_functions)

        converted_sil_progs = self._convert_silver_elements(sil_progs,
                                                            all_used_names, ctx)
        s_domains, s_predicates, s_functions, s_methods = converted_sil_progs
        domains += s_domains
        predicates += s_predicates
        functions += s_functions
        methods += s_methods

        prog = self.viper.Program(domains, fields, functions, predicates,
                                  methods, self.no_position(ctx),
                                  self.no_info(ctx))
        return prog



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Pytest plugin for further test configuration.

See http://doc.pytest.org/en/latest/writing_plugins.html for documentation.
"""
import os
import pytest

from nagini_translation.lib import config
from nagini_translation.tests import _JVM as jvm
from nagini_translation.verifier import ViperVerifier
from typing import List


_TRANSLATION_TEST_FUNCTION_NAME = 'test_translation'
_VERIFICATION_TEST_FUNCTION_NAME = 'test_verification'

_TRANSLATION_TESTS_SUFFIX = 'translation'
_VERIFICATION_TESTS_SUFFIX = 'verification'

_FUNCTIONAL_TESTS_DIR = 'tests/functional/'
_SIF_TESTS_DIR = 'tests/sif/'
_IO_TESTS_DIR = 'tests/io/'
_OBLIGATIONS_TESTS_DIR = 'tests/obligations/'
_ARP_TESTS_DIR = 'tests/arp/'


class PyTestConfig:
    """Class that holds the configuration for tests."""
    def __init__(self):
        self.translation_test_dirs = []
        self.verification_test_dirs = []
        self.single_test = None
        self.verifiers = []

        self.init_from_config_file()

    def init_from_config_file(self):
        test_config = config.test_config

        for test in test_config.tests:
            self.add_test(test)

        for verifier in test_config.verifiers:
            self.add_verifier(verifier)

    def add_test(self, test: str):
        if test == 'functional':
            self._add_test_dir(_FUNCTIONAL_TESTS_DIR)
        elif test == 'sif':
            self._add_test_dir(_SIF_TESTS_DIR)
        elif test == 'io':
            self._add_test_dir(_IO_TESTS_DIR)
        elif test == 'obligations':
            self._add_test_dir(_OBLIGATIONS_TESTS_DIR)
        elif test == 'arp':
            self._add_test_dir(_ARP_TESTS_DIR)
        else:
            print('Unrecognized test set.')

    def _add_test_dir(self, path: str):
        self.translation_test_dirs.append(os.path.join(
            path, _TRANSLATION_TESTS_SUFFIX))
        self.verification_test_dirs.append(os.path.join(
            path, _VERIFICATION_TESTS_SUFFIX))

    def clear_tests(self):
        self.translation_test_dirs = []
        self.verification_test_dirs = []

    def add_verifier(self, verifier: str):
        if verifier == 'silicon':
            self.verifiers.append(ViperVerifier.silicon)
        elif verifier == 'carbon':
            self.verifiers.append(ViperVerifier.carbon)
        else:
            print('Unrecognized verifier.')

    def clear_verifiers(self):
        self.verifiers = []

_pytest_config = PyTestConfig()


def _test_files(test_dir: str) -> List[str]:
    result = []
    for root, dir_names, file_names in os.walk(test_dir, topdown=True):
        if 'tests' in file_names:
            # tests file lists all tests in this directory, so we read
            # its contents and do not proceed deeper.
            with open(os.path.join(root, 'tests')) as fp:
                for file_name in fp:
                    result.append(os.path.join(root, file_name.strip()))
            dir_names.clear()
            continue
        if 'resources' in dir_names:
            # Skip resources directory.
            dir_names.remove('resources')
        for file_name in file_names:
            if file_name.endswith('.py'):
                result.append(os.path.join(root, file_name))
    result = [path for path in sorted(result)
              if path not in config.test_config.ignore_tests]
    return result


def pytest_addoption(parser: 'pytest.config.Parser'):
    """Command line options for the test runner."""
    # Preferably, we could specify the tests and verifiers as a list, but
    # unfortunately, pytest_parser.addoption does not play well with
    # action='append'.
    parser.addoption('--single-test', dest='single_test', action='store', default=None)
    parser.addoption('--all-tests', dest='all_tests', action='store_true')
    parser.addoption('--functional', dest='functional', action='store_true')
    parser.addoption('--sif', dest='sif', action='store_true')
    parser.addoption('--io', dest='io', action='store_true')
    parser.addoption('--obligations', dest='obligations', action='store_true')
    parser.addoption('--arp', dest='arp', action='store_true')
    parser.addoption('--all-verifiers', dest='all_verifiers',
                     action='store_true')
    parser.addoption('--silicon', dest='silicon', action='store_true')
    parser.addoption('--carbon', dest='carbon', action='store_true')


def pytest_configure(config: 'pytest.config.Config'):
    """Adds command line arguments to the PyTestConfig object."""
    # Setup tests.
    tests = []
    if config.option.all_tests:
        tests = ['functional', 'sif', 'io', 'obligations', 'arp']
    else:
        if config.option.functional:
            tests.append('functional')
        if config.option.sif:
            tests.append('sif')
        if config.option.io:
            tests.append('io')
        if config.option.obligations:
            tests.append('obligations')
        if config.option.arp:
            tests.append('arp')
    if tests:
        # Overwrite config file options.
        _pytest_config.clear_tests()
        for test in tests:
            _pytest_config.add_test(test)
        if 'sif' in tests:
            if not jvm.is_known_class(jvm.viper.silver.sif.SIFReturnStmt):
                pytest.exit('Viper SIF extension not avaliable on the classpath.')
    elif config.option.single_test:
        _pytest_config.clear_tests()
        _pytest_config.single_test = config.option.single_test
    if not _pytest_config.translation_test_dirs and not _pytest_config.single_test:
        # Default: all tests that are available, SIF tests only if the extension is
        # present.
        tests = ['functional', 'io', 'obligations']
        if jvm.is_known_class(jvm.viper.silver.sif.SIFReturnStmt):
            tests.append('sif')
        if jvm.is_known_class(jvm.viper.silver.plugin.ARPPlugin):
            tests.append('arp')
        for test in tests:
            _pytest_config.add_test(test)
    # Setup verifiers.
    verifiers = []
    if config.option.all_verifiers:
        verifiers = ['silicon', 'carbon']
    else:
        if config.option.silicon:
            verifiers.append('silicon')
        if config.option.carbon:
            verifiers.append('carbon')
    if verifiers:
        # Overwrite config file options.
        _pytest_config.clear_verifiers()
        for verifier in verifiers:
            _pytest_config.add_verifier(verifier)
    if not _pytest_config.verifiers:
        # Default: all available verifiers.
        verifiers = []
        if jvm.is_known_class(jvm.viper.silicon.SiliconRunner):
            verifiers.append('silicon')
        if jvm.is_known_class(jvm.viper.carbon.Carbon):
            verifiers.append('carbon')
        if not verifiers:
            pytest.exit('No backend verifiers avaliable on the classpath.')
        for verifier in verifiers:
            _pytest_config.add_verifier(verifier)


def pytest_generate_tests(metafunc: 'pytest.python.Metafunc'):
    """Parametrizes test functions based on the config."""
    func_name = metafunc.function.__name__
    test_files = []
    reload_triggers = set()
    params = []
    if func_name == _TRANSLATION_TEST_FUNCTION_NAME:
        for test_dir in _pytest_config.translation_test_dirs:
            files = _test_files(test_dir)
            test_files.extend(files)
            reload_triggers.add(files[0])
        if _pytest_config.single_test and 'translation' in _pytest_config.single_test:
            test_files.append(_pytest_config.single_test)
        for file in test_files:
            sif = 'sif' in file
            reload_resources = file in reload_triggers
            arp = 'arp' in file
            params.append((file, sif, reload_resources, arp))
        metafunc.parametrize('path,sif,reload_resources,arp', params)
    elif func_name == _VERIFICATION_TEST_FUNCTION_NAME:
        for test_dir in _pytest_config.verification_test_dirs:
            files = _test_files(test_dir)
            test_files.extend(files)
            reload_triggers.add(files[0])
        if _pytest_config.single_test and 'verification' in _pytest_config.single_test:
            test_files.append(_pytest_config.single_test)
        for file in test_files:
            sif = 'sif' in file
            reload_resources = file in reload_triggers
            arp = 'arp' in file
            params.extend([(file, verifier, sif, reload_resources, arp) for verifier
                           in _pytest_config.verifiers])
        metafunc.parametrize('path,verifier,sif,reload_resources,arp', params)
    else:
        pytest.exit('Unrecognized test function.')



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""
A singleton configuration object.

>>> from nagini_translation.lib import config
>>> config.classpath is None
False
>>> config.z3_path is None
False
>>> config.mypy_path is None
False

"""


import configparser
import glob
import os
import sys

from typing import List


class SectionConfig:
    """A base class for configuration sections."""

    def __init__(self, config, section_name) -> None:
        self.config = config
        if section_name not in self.config:
            self.config[section_name] = {}
        self._info = self.config[section_name]


class ObligationConfig(SectionConfig):
    """Obligation translation configuration."""

    def __init__(self, config) -> None:
        super().__init__(config, 'Obligations')

    @property
    def disable_all(self):
        """Disable all obligation related checks."""
        return self._info.getboolean('disable_all', False)

    @property
    def disable_measure_check(self):
        """Replace obligation measure checks with ``True``."""
        return self._info.getboolean(
            'disable_measure_check', self.disable_all)

    @property
    def disable_measures(self):
        """Completely disable obligation measures."""
        return self._info.getboolean(
            'disable_measures', self.disable_all)

    @property
    def disable_method_body_leak_check(self):
        """Disable leak check at the end of method body."""
        return self._info.getboolean(
            'disable_method_body_leak_check', self.disable_all)

    @property
    def disable_loop_body_leak_check(self):
        """Disable leak check at the end of loop body."""
        return self._info.getboolean(
            'disable_loop_body_leak_check', self.disable_all)

    @property
    def disable_call_context_leak_check(self):
        """Disable leak check at the caller side."""
        return self._info.getboolean(
            'disable_call_context_leak_check', self.disable_all)

    @property
    def disable_loop_context_leak_check(self):
        """Disable leak check at the loop surrounding context."""
        return self._info.getboolean(
            'disable_loop_context_leak_check', self.disable_all)

    @property
    def disable_termination_check(self):
        """Disable all termination checks.

        Also replaces all ``MustTerminate`` with ``True``.
        """
        return self._info.getboolean(
            'disable_termination_check', self.disable_all)

    @property
    def disable_must_invoke(self):
        """Replace all ``token`` with ``ctoken``."""
        return self._info.getboolean(
            'disable_must_invoke', self.disable_all)

    @property
    def disable_waitlevel_check(self):
        """Disable all waitlevel checks."""
        return self._info.getboolean(
            'disable_waitlevel_check', self.disable_all)


class TestConfig(SectionConfig):
    """Testing configuration."""

    def __init__(self, config) -> None:
        super().__init__(config, 'Tests')

        ignore_tests_value = self._info.get('ignore_tests')
        if not ignore_tests_value:
            self.ignore_tests = set([])
        else:
            patterns = ignore_tests_value.strip().splitlines()
            self.ignore_tests = set([i for pattern in patterns for i in glob.glob(pattern)])

        verifiers_value = self._info.get('verifiers')
        if not verifiers_value:
            self.verifiers = []
        else:
            self.verifiers = verifiers_value.strip().split()

        tests_value = self._info.get('tests')
        if not tests_value:
            self.tests = []
        else:
            self.tests = tests_value.strip().split()


class FileConfig:
    """Configuration stored in the config file."""

    def __init__(self, config_file) -> None:
        self.config = configparser.ConfigParser()
        self.config.read(config_file)
        self.obligation_config = ObligationConfig(self.config)
        self.test_config = TestConfig(self.config)


def _construct_classpath(verifier : str = None):
    """ Contstructs JAVA classpath.

    First tries environment variables ``VIPERJAVAPATH``, ``SILICONJAR``
    and ``CARBONJAR``. If they are undefined, then tries to use OS
    specific locations.
    """

    viper_java_path = os.environ.get('VIPERJAVAPATH')
    silicon_jar = os.environ.get('SILICONJAR')
    carbon_jar = os.environ.get('CARBONJAR')
    arpplugin_jar = os.environ.get('ARPPLUGINJAR')

    if viper_java_path:
        return viper_java_path

    if silicon_jar or carbon_jar:
        return os.pathsep.join(
            jar for jar, v in ((silicon_jar, 'carbon'),
                               (carbon_jar, 'silicon'),
                               (arpplugin_jar, 'arpplugin'))
            if jar and v != verifier)

    if sys.platform.startswith('linux'):
        if os.path.isdir('/usr/lib/viper'):
            # Check if we have Viper installed via package manager.
            return os.pathsep.join(
                glob.glob('/usr/lib/viper/*.jar'))

    return ''


def _get_boogie_path():
    """ Tries to detect path to Boogie executable.

    First tries the environment variable ``BOOGIE_EXE``. If it is not
    defined, then checks the OS specific directory. On Ubuntu returns a
    path only if it also finds a mono installation.
    """

    boogie_exe = os.environ.get('BOOGIE_EXE')
    if boogie_exe:
        return boogie_exe

    if sys.platform.startswith('linux'):
        if (os.path.exists('/usr/bin/boogie') and
            os.path.exists('/usr/bin/mono')):
            return '/usr/bin/boogie'


def _get_z3_path():
    """ Tries to detect path to Z3 executable.

    First tries the environment variable ``Z3_EXE``. If it is not
    defined, then checks the OS specific directories.
    """

    z3_exe = os.environ.get('Z3_EXE')
    if z3_exe:
        return z3_exe

    if sys.platform.startswith('linux'):
        if os.path.exists('/usr/bin/viper-z3'):
            # First check if we have Z3 installed together with Viper.
            return '/usr/bin/viper-z3'
        if os.path.exists('/usr/bin/z3'):
            return '/usr/bin/z3'


def _get_mypy_path():
    """ Construct MYPYPATH.

    If MYPYPATH environment variable is not defined, then sets it to
    ``nagini-contracts`` directory.
    """

    mypy_path = os.environ.get('MYPYPATH')
    if not mypy_path:
        import nagini_contracts
        mypy_path = os.path.dirname(os.path.dirname(
            nagini_contracts.__file__))
    return mypy_path


def _get_mypy_dir():
    """ Construct MYPYDIR.

    If MYPYDIR environment variable is not defined, just returns None.
    """

    mypy_dir = os.environ.get('MYPYDIR')
    if mypy_dir:
        return os.path.dirname(mypy_dir)
    return None


def set_verifier(v: str):
    global classpath
    not_set_by_arg = classpath == _construct_classpath()
    if not_set_by_arg:
        classpath = _construct_classpath(v)


mypy_dir = _get_mypy_dir()
"""
Mypy executable dir. Initialized by calling
:py:func:`_get_mypy_dir`.
"""


classpath = _construct_classpath()
"""
JAVA class path. Initialized by calling
:py:func:`_construct_classpath`.
"""


boogie_path = _get_boogie_path()
"""
Path to Boogie executable. Initialized by calling
:py:func:`_get_boogie_path`.
"""


z3_path = _get_z3_path()
"""
Path to Z3 executable. Initialized by calling :py:func:`_get_z3_path`.
"""


mypy_path = _get_mypy_path()
"""
MYPY search path. Initialized by calling :py:func:`_get_mypy_path`.
"""


file_config = FileConfig('nagini.cfg')
"""
Configuration read from ``nagini.cfg`` file.
"""


obligation_config = file_config.obligation_config
"""
Obligation configuration.
"""


test_config = file_config.test_config
"""
Test configuration.
"""


__all__ = (
    'classpath',
    'boogie_path',
    'z3_path',
    'mypy_path',
    'mypy_dir',
    'obligation_config',
    'set_verifier',
)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

from typing import List, Optional

from nagini_translation.lib.typedefs import (
    Expr, Info, Position, Seqn, Stmt, Var, VarAssign
)
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.sif.lib.util import (
    in_postcondition_of_dyn_bound_call, in_override_check
)


class ViperASTExtended(ViperAST):
    """
    Extends the Viper AST, to provide access to extension nodes.
    """

    def __init__(self, jvm, java, scala, viper, sourcefile):
        super().__init__(jvm, java, scala, viper, sourcefile)
        self.ast_extensions = viper.silver.sif
        self.all_low_methods = set()
        self.preserves_low_methods = set()
        self.equality_comp_functions = set()
        self.ctx = None
        self.type_factory = None

    def is_extension_available(self) -> bool:
        """
        Checks if the extended AST is available, i.e., the SIF AST extension is on the
        Java classpath.
        """
        return self.jvm.is_known_class(self.ast_extensions.SIFReturnStmt)

    def Return(self, expr: Optional[Expr], res_var: Optional[Var], position: Position,
               info: Info):
        expr_opt = self.scala.Some(expr) if expr is not None else self.none
        res_var_opt = self.scala.Some(res_var) if res_var is not None else self.none
        return self.ast_extensions.SIFReturnStmt(expr_opt, res_var_opt,
                                                 position, info, self.NoTrafos)

    def Break(self, position: Position, info: Info):
        return self.ast_extensions.SIFBreakStmt(position, info, self.NoTrafos)

    def Continue(self, position: Position, info: Info):
        return self.ast_extensions.SIFContinueStmt(position, info, self.NoTrafos)

    def Skip(self) -> Seqn:
        return self.Seqn([], self.NoPosition, self.NoInfo)

    def Goto(self, name: str, position: Position, info: Info) -> Stmt:
        """ Don't add the gotos, not needed in extended AST. """
        return self.Skip()

    def Label(self, name: str, position: Position, info: Info) -> Stmt:
        """ Don't add the labels, not needed in extended AST. """
        return self.Skip()

    def Raise(self, assignment: Optional[VarAssign], position: Position, info: Info):
        ass_opt = self.scala.Some(assignment) if assignment is not None else self.none
        return self.ast_extensions.SIFRaiseStmt(ass_opt, position, info, self.NoTrafos)

    def SIFExceptionHandler(self, err_var: Var, exception: Expr, handler: Seqn):
        return self.ast_extensions.SIFExceptionHandler(
            err_var, exception, handler, self.NoPosition, self.NoInfo, self.NoTrafos)

    def Try(self, body: Seqn, catch_blocks: List['silver.sif.SIFExceptionHandler'],
            else_block: Optional[Seqn], finally_block: Optional[Seqn],
            position: Position, info: Info):
        catch_blocks_seq = self.to_seq(catch_blocks)
        else_opt = self.scala.Some(else_block) if else_block is not None else self.none
        fin_opt = (self.scala.Some(finally_block)
                   if finally_block is not None else self.none)
        return self.ast_extensions.SIFTryCatchStmt(
            body, catch_blocks_seq, else_opt, fin_opt, position, info, self.NoTrafos)

    def Low(self, expr: Expr, comp: Optional[str], position: Position, info: Info):
        if comp:
            self.used_names.add(comp)
            comp_opt = self.scala.Some(comp)
        else:
            comp_opt = self.none
        return self.ast_extensions.SIFLowExp(
            expr, comp_opt, position, info, self.NoTrafos)

    def LowEvent(self, position: Position, info: Info):
        return self.ast_extensions.SIFLowEventExp(position, info, self.NoTrafos)

    def Declassify(self, expr: Expr, position: Position, info: Info):
        return self.ast_extensions.SIFDeclassifyStmt(expr, position, info, self.NoTrafos)

    def InlinedCall(self, stmts: Seqn, position: Position, info: Info):
        return self.ast_extensions.SIFInlinedCallStmt(stmts, position, info, self.NoTrafos)

    def AssertNoException(self, position: Position, info: Info):
        return self.ast_extensions.SIFAssertNoException(position, info, self.NoTrafos)

    def TerminatesSif(self, cond: Expr, position: Position, info: Info):
        return self.ast_extensions.SIFTerminatesExp(cond, position, info, self.NoTrafos)

    def PredicateAccessPredicate(self, loc, perm, position, info):
        if self.ctx and self.type_factory:
            dyn_check = in_postcondition_of_dyn_bound_call(self.type_factory, self.ctx)
            if dyn_check:
                info = self.ConsInfo(
                    self.SIFDynCheckInfo([], dyn_check, in_override_check(self.ctx)),
                    info)
        return super().PredicateAccessPredicate(loc, perm, position, info)

    def SIFInfo(self, comments: List[str],
                continue_unaware: bool = False,
                obligation_var: bool = False) -> 'silver.sif.SIFInfo':
        return self.ast_extensions.SIFInfo(
            self.to_seq(comments), continue_unaware, obligation_var)

    def SIFDynCheckInfo(self, comments: List[str],
                        dyn_check: Expr,
                        dyn_check_only: bool = False) -> 'silver.sif.SIFDynCheckInfo':
        return self.ast_extensions.SIFDynCheckInfo(self.to_seq(comments), dyn_check,
                                                   dyn_check_only)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Code for working with obligation measures."""


from typing import List

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.context import Context
from nagini_translation.lib.config import obligation_config
from nagini_translation.lib.program_nodes import (
    PythonVar,
)
from nagini_translation.lib.typedefs import (
    Stmt,
)


class MeasureMap:
    """Abstraction over map from obligation references to their measures."""

    def __init__(self, measure_map_var: PythonVar) -> None:
        self._map_var = measure_map_var
        self._measure_domain = sil.Domain('Measure$')
        self._measure_domain.declare_function('Measure$create')

    @property
    def _measure_var_type(self) -> sil.Type:
        return sil.SeqType(self._measure_domain.get_type())

    def _create_measure(
            self, guard: sil.BoolExpression, key: sil.RefExpression,
            value: sil.IntExpression) -> sil.DomainFuncApp:
        return self._measure_domain.call_function(
            'Measure$create', [guard, key, value])

    def check(self, reference: sil.RefExpression,
              value: sil.IntExpression) -> sil.BoolExpression:
        """Generate a check if current value is smaller than in map."""
        if (obligation_config.disable_measures or
                obligation_config.disable_measure_check):
            return sil.TrueLit()
        args = [
            sil.CallArg('map', self._measure_var_type,
                        sil.AnyVar(self._map_var)),
            sil.CallArg('key', sil.REF, reference),
            sil.CallArg('value', sil.INT, value),
        ]
        return sil.BoolCall('Measure$check', args)

    def get_var(self) -> PythonVar:
        """Return a variable representing the measure map."""
        return self._map_var

    def initialize(
            self, obligation_instances: List['GuardedObligationInstance'],
            translator: 'AbstractTranslator', ctx: Context,
            overriding_check: bool = False) -> List[Stmt]:
        """Construct a list of statements that initialize measure map.

        If ``overriding_check`` is ``True``, then adds to measures ``1``
        to allow overridden method to have the same measure.
        """
        position = translator.no_position(ctx)
        info = translator.no_info(ctx)
        measures = []
        for instance in obligation_instances:
            if instance.obligation_instance.is_fresh():
                continue
            guard_expression = instance.create_guard_expression()
            value = instance.obligation_instance.get_measure()
            if overriding_check:
                value = sil.Inc(value)
            measure = self._create_measure(
                guard_expression,
                instance.obligation_instance.get_target(), value)
            measures.append(measure)
        assign = sil.Assign(
            self._map_var,
            sil.PSeq(self._measure_domain.get_type(), measures))
        statement = assign.translate(translator, ctx, position, info)
        return [statement]



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Classes for performing well-formedness checks of IO stuff."""


import ast

from typing import List

from nagini_translation.lib.util import InvalidProgramException


def _raise_error(node, error_type) -> None:
    """Raise error exception."""
    raise InvalidProgramException(
        node,
        'invalid.io_operation.body.' + error_type,
    )


class IOOperationBodyChecker(ast.NodeVisitor):
    """Perform some simple well formedness checks for IOOperation body.

    Performed checks:

    1.  ``IOExists`` are not nested.
    2.  Existential variables are defined before their use.
    """

    def __init__(
            self, body: ast.Expr,
            results: List['lib.program_nodes.PythonVar'],
            io_existentials: List['lib.program_nodes.PythonVarCreator'],
            module: 'lib.program_nodes.PythonModule',
            translator: 'Translator') -> None:
        super().__init__()
        self._body = body
        self._results = {var.name for var in results}
        self._existentials = dict(
            (var.name, var) for var in io_existentials)
        self._undefined_existentials = set(self._existentials.keys())
        self._module = module
        self._translator = translator
        self._counter = 0

    def check(self) -> None:
        """Check that body is well formed."""
        self.visit(self._body)

    def visit_Name(self, node: ast.Name) -> None:
        """Check that only allowed names are used in the body."""
        name = node.id
        if name.startswith('IOExists'):
            _raise_error(node, 'ioexists')
        if name == 'Acc':
            _raise_error(node, 'non_pure')
        if name in self._undefined_existentials:
            _raise_error(node, 'use_of_undefined_existential')

    def visit_Call(self, node: ast.Call) -> None:
        """Check IO operation use and define existential variables."""
        containers = [self._module]
        containers.extend(self._module.get_included_modules())
        target = None
        if not isinstance(node.func, ast.Call):
            target = self._translator.get_target(node, containers, None)
        if target and target.__class__.__name__ == 'PythonIOOperation':
            operation = target
            parameter_count = len(operation.get_parameters())
            results = node.args[parameter_count:]
            for i, arg in enumerate(results):
                if not isinstance(arg, ast.Name):
                    _raise_error(node, 'not_variable_in_result_position')
                elif (arg.id not in self._existentials and
                      arg.id not in self._results):
                    _raise_error(node, 'variable_not_existential')
                elif arg.id in self._undefined_existentials:
                    self._undefined_existentials.remove(arg.id)
                    var = self._existentials[arg.id]
                    var.set_defining_info(self._counter, node,
                                          operation.get_results()[i])
                    self._counter += 1
        self.visit(node.func)
        for arg in node.args:
            self.visit(arg)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Conversion rules from Silver level errors to Nagini errors."""


from typing import Dict, Tuple


class Rules(Dict[Tuple[str, str], Tuple[str, str]]):
    """Error conversion rules."""


TERMINATION_CHECK_MEASURE_NON_POSITIVE = {
    ('assert.failed', 'assertion.false'):
        ('termination_check.failed', 'termination_measure.non_positive')
}
TERMINATION_CHECK_MEASURE_NON_DECREASING = {
    ('assert.failed', 'assertion.false'):
        ('termination_check.failed', 'measure.non_decreasing')
}
TERMINATION_CHECK_GAP_ENABLED = {
    ('assert.failed', 'assertion.false'):
        ('termination_check.failed', 'gap.enabled')
}
TERMINATION_CHECK_CHILD_TERMINATION_NOT_IMPLIED = {
    ('assert.failed', 'assertion.false'):
        ('termination_check.failed', 'child_termination.not_implied')
}

OBLIGATION_MEASURE_NON_POSITIVE = {
    ('call.precondition', 'assertion.false'):
        ('call.precondition', 'obligation_measure.non_positive')
}
OBLIGATION_LOOP_TERMINATION_PROMISE_FAIL = {
    ('assert.failed', 'assertion.false'):
        ('leak_check.failed', 'must_terminate.loop_promise_not_kept')
}
OBLIGATION_CALL_LEAK_CHECK_FAIL = {
    ('call.precondition', 'assertion.false'):
        ('leak_check.failed', 'caller.has_unsatisfied_obligations')
}
OBLIGATION_BODY_LEAK_CHECK_FAIL = {
    ('postcondition.violated', 'assertion.false'):
        ('leak_check.failed', 'method_body.leaks_obligations')
}
OBLIGATION_LOOP_CONTEXT_LEAK_CHECK_FAIL = {
    ('invariant.not.established', 'assertion.false'):
        ('leak_check.failed', 'loop_context.has_unsatisfied_obligations')
}
OBLIGATION_LOOP_BODY_LEAK_CHECK_FAIL = {
    ('invariant.not.preserved', 'assertion.false'):
        ('leak_check.failed', 'loop_body.leaks_obligations')
}
OBLIGATION_LOOP_MEASURE_NON_POSITIVE = {
    ('invariant.not.established', 'assertion.false'):
        ('invariant.not.established', 'obligation_measure.non_positive'),
    ('invariant.not.preserved', 'assertion.false'):
        ('invariant.not.preserved', 'obligation_measure.non_positive'),
    ('not.wellformed', 'insufficient.permission'):
        ('not.wellformed', 'loop_condition.not_framed_for_obligation_use'),
}
LOCAL_VARIABLE_NOT_DEFINED = {
    ('application.precondition', 'assertion.false'):
        ('expression.undefined', 'undefined.local.variable')
}
GLOBAL_NAME_NOT_DEFINED = {
    ('application.precondition', 'assertion.false'):
        ('expression.undefined', 'undefined.global.name')
}
DEPENDENCIES_NOT_DEFINED = {
    ('application.precondition', 'assertion.false'):
        ('expression.undefined', 'missing.dependencies')
}
THREAD_CREATION_ARG_TYPE = {
    ('assert.failed', 'assertion.false'):
        ('thread.creation.failed', 'invalid.argument.type')
}
THREAD_CREATION_GROUP_NONE = {
    ('assert.failed', 'assertion.false'):
        ('thread.creation.failed', 'assertion.false')
}
THREAD_START_METHOD_UNLISTED = {
    ('assert.failed', 'assertion.false'):
        ('thread.start.failed', 'method.not.listed')
}
THREAD_START_PERMISSION = {
    ('exhale.failed', 'insufficient.permission'):
        ('thread.start.failed', 'missing.start.permission')
}
THREAD_START_PRECONDITION = {
    ('exhale.failed', 'insufficient.permission'):
        ('thread.start.failed', 'insufficient.permission'),
    ('inhale.failed', 'insufficient.permission'):
        ('thread.start.failed', 'insufficient.permission'),
    ('exhale.failed', 'assertion.false'):
        ('thread.start.failed', 'assertion.false')
}
THREAD_JOIN_WAITLEVEL = {
    ('assert.failed', 'assertion.false'):
        ('thread.join.failed', 'wait.level.invalid')
}
THREAD_JOIN_JOINABLE = {
    ('assert.failed', 'assertion.false'):
        ('thread.join.failed', 'thread.not.joinable')
}
INHALE_TO_CALL = {
    ('inhale.failed', 'insufficient.permission'):
        ('call.precondition', 'insufficient.permission')
}
LOCK_RELEASE_INVARIANT = {
    ('fold.failed', 'insufficient.permission'):
        ('lock.invariant.not.established', 'insufficient.permission'),
    ('fold.failed', 'assertion.false'):
        ('lock.invariant.not.established', 'assertion.false')
}

__all__ = (
    'Rules',
    'TERMINATION_CHECK_MEASURE_NON_POSITIVE',
    'TERMINATION_CHECK_MEASURE_NON_DECREASING',
    'TERMINATION_CHECK_GAP_ENABLED',
    'TERMINATION_CHECK_CHILD_TERMINATION_NOT_IMPLIED',
    'OBLIGATION_MEASURE_NON_POSITIVE',
    'OBLIGATION_LOOP_TERMINATION_PROMISE_FAIL',
    'OBLIGATION_CALL_LEAK_CHECK_FAIL',
    'OBLIGATION_BODY_LEAK_CHECK_FAIL',
    'OBLIGATION_LOOP_CONTEXT_LEAK_CHECK_FAIL',
    'OBLIGATION_LOOP_BODY_LEAK_CHECK_FAIL',
    'OBLIGATION_LOOP_MEASURE_NON_POSITIVE',
    'LOCAL_VARIABLE_NOT_DEFINED',
    'GLOBAL_NAME_NOT_DEFINED',
    'DEPENDENCIES_NOT_DEFINED',
)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""
Monkey-patches the mypy parser to add column information to the mypy AST.
More precisely, we monkey-patch the transformer that takes the normal
Python AST and turns it into a mypy AST to add, for every node, the column
information (if any) from the original node to the 'column' field of the new
node.

We need the column information because otherwise we have no way to know
the precise types of different references to the same variable in the same line
if they are different (which they may be because of isinstance checks).
"""
import mypy.fastparse

from functools import wraps


def with_column(f):
    """
    Decorator for functions belonging to the translation from Python AST to
    mypy AST. Adds the column information from the Python node to the column
    field of the mypy node.
    """
    @wraps(f)
    def wrapper(self, ast):
        node = f(self, ast)
        if hasattr(ast, 'col_offset'):
            node.column = ast.col_offset
        return node
    return wrapper


# Monkey-patch mypy's AST converter to add column information
ASTConverter = mypy.fastparse.ASTConverter
for name in dir(ASTConverter):
    m = getattr(ASTConverter, name)
    if callable(m) and name.startswith('visit_'):
        setattr(ASTConverter, name, with_column(m))



#!/usr/bin/env python3

from setuptools import setup, find_packages


setup(
        name='nagini',
        version='0.8',
        author='Viper Team',
        author_email='viper@inf.ethz.ch',
        license='MPL-2.0',
        packages=find_packages('src'),
        package_dir={'': 'src'},
        requires=[
            'distribute',
            ],
        install_requires=[
            'mypy-lang',
            'toposort',
            'jpype1',
            'astunparse',
            'typed-ast',
            'pyzmq',
            ],
        entry_points = {
             'console_scripts': [
                 'nagini = nagini_translation.main:main',
                 ]
             },
        url='http://www.pm.inf.ethz.ch/research/nagini.html',
        description='Static verifier for Python 3, based on Viper.',
        # Full list of classifiers could be found at:
        # http://pypi.python.org/pypi?%3Aaction=list_classifiers
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3 :: Only',
            'Topic :: Software Development',
            ],
        )



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Public interface to obligation translator."""


import ast

from typing import List, Tuple, Union

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.context import Context
from nagini_translation.lib.jvmaccess import JVM
from nagini_translation.lib.program_nodes import (
    PythonMethod,
    PythonVar,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Field,
    Method,
    Position,
    Predicate,
    Stmt,
    StmtsAndExpr,
    VarDecl,
)
from nagini_translation.lib.typeinfo import TypeInfo
from nagini_translation.lib.util import (
    get_func_name,
    InvalidProgramException,
    UnsupportedException,
)
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translators.abstract import TranslatorConfig
from nagini_translation.translators.common import CommonTranslator
from nagini_translation.translators.obligation.loop import (
    LoopObligationTranslator,
)
from nagini_translation.translators.obligation.manager import (
    ObligationManager,
)
from nagini_translation.translators.obligation.method import (
    MethodObligationTranslator,
)
from nagini_translation.translators.obligation.obligation_info import (
    PythonMethodObligationInfo,
)
from nagini_translation.translators.obligation.waitlevel import (
    WaitLevelTranslator,
)


class ObligationTranslator(CommonTranslator):
    """Translator for obligations."""

    def __init__(self, config: TranslatorConfig, jvm: JVM, source_file: str,
                 type_info: TypeInfo, viper_ast: ViperAST) -> None:
        super().__init__(config, jvm, source_file, type_info, viper_ast)
        self._obligation_manager = ObligationManager()
        self._method_translator = MethodObligationTranslator(
            config, jvm, source_file, type_info, viper_ast,
            self._obligation_manager)
        self._loop_translator = LoopObligationTranslator(
            config, jvm, source_file, type_info, viper_ast,
            self._obligation_manager)
        self._wait_level_translator = WaitLevelTranslator(
            config, jvm, source_file, type_info, viper_ast,
            self._obligation_manager)

    def enter_loop_translation(
            self, node: Union[ast.While, ast.For], ctx: Context,
            err_var: PythonVar = None) -> None:
        """Update context with info needed to translate loop."""
        self._loop_translator.enter_loop_translation(node, ctx, err_var)

    def leave_loop_translation(self, ctx: Context) -> None:
        """Remove loop translation info from context."""
        self._loop_translator.leave_loop_translation(ctx)

    def create_while_node(
            self, ctx: Context, cond: Expr,
            invariants: List[Expr],
            local_vars: List[VarDecl],
            body: Stmt, node: Union[ast.While, ast.For]) -> List[Stmt]:
        """Construct a while loop AST node with obligation stuff."""
        return self._loop_translator.create_while_node(
            ctx, cond, invariants, local_vars, body, node)

    def translate_obligation_contractfunc_call(
            self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """Translate a call to obligation contract function."""
        func_name = get_func_name(node)
        if func_name == 'MustTerminate':
            return self._translate_must_terminate(node, ctx)
        elif func_name == 'MustRelease':
            return self._translate_must_release(node, ctx)
        elif func_name == 'WaitLevel':
            raise InvalidProgramException(
                node, 'invalid.wait_level.use')
        elif func_name == 'Level':
            raise InvalidProgramException(
                node, 'invalid.level.use')
        else:
            raise UnsupportedException(
                node, 'Unsupported contract function.')

    def translate_must_invoke_token(
            self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """Translate a call to ``token``."""
        if ctx.obligation_context.is_translating_loop():
            return self._loop_translator.translate_must_invoke(node, ctx)
        else:
            return self._method_translator.translate_must_invoke(node, ctx)

    def translate_must_invoke_ctoken(
            self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """Translate a call to ``ctoken``."""
        if ctx.obligation_context.is_translating_loop():
            return self._loop_translator.translate_may_invoke(
                node, ctx)
        elif ctx.obligation_context.is_translating_posts:
            if ctx.actual_function.name != 'Gap':
                raise InvalidProgramException(
                    node, 'invalid.postcondition.ctoken_not_allowed')
            else:
                return self._method_translator.translate_may_invoke(
                    node, ctx)
        else:
            return self._method_translator.translate_may_invoke(
                node, ctx)

    def is_wait_level_comparison(self, node: ast.Compare,
                                 ctx: Context) -> bool:
        """Check if we are comparing with ``WaitLevel``."""
        return self._wait_level_translator.is_wait_level_comparison(
            node, ctx)

    def translate_wait_level_comparison(self, node: ast.Compare,
                                        ctx: Context) -> StmtsAndExpr:
        """Translate comparison with ``WaitLevel``."""
        return self._wait_level_translator.translate_wait_level_comparison(
            node, ctx)

    def create_level_call(self, expr: sil.RefExpression) -> sil.PermExpression:
        return self._wait_level_translator.create_level_call(expr)

    def create_level_below(
            self, expr: sil.PermExpression,
            residue_level_var: sil.PermExpression,
            ctx: Context) -> sil.BoolExpression:
        return self._wait_level_translator.create_level_below(
            expr, residue_level_var, ctx)

    def initialize_current_wait_level(
            self, current_wait_level: sil.PermExpression,
            residue_level_var: sil.PermExpression,
            ctx: Context) -> sil.BoolExpression:
        return self._wait_level_translator.initialize_current_wait_level(
            current_wait_level, residue_level_var, ctx)

    def _translate_must_terminate(
            self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """Translate a call to ``MustTerminate``."""
        if ctx.obligation_context.is_translating_loop():
            return self._loop_translator.translate_must_terminate(node, ctx)
        elif ctx.obligation_context.is_translating_posts:
            raise InvalidProgramException(
                node, 'obligation.must_terminate.in_postcondition')
        else:
            return self._method_translator.translate_must_terminate(node, ctx)

    def _translate_must_release(
            self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """Translate a call to ``MustRelease``."""
        if ctx.obligation_context.is_translating_loop():
            return self._loop_translator.translate_must_release(node, ctx)
        else:
            return self._method_translator.translate_must_release(node, ctx)

    def get_obligation_preamble(
            self,
            ctx: Context) -> Tuple[List[Predicate], List[Field]]:
        """Construct obligation preamble.

        To track each obligation we use predicates which are defined in
        this preamble.
        """
        predicates = self._obligation_manager.create_predicates(self)
        fields = self._obligation_manager.create_fields(self)
        return predicates, fields

    def create_method_node(     # pylint: disable=too-many-arguments
            self, ctx: Context, name: str,
            args: List[VarDecl], returns: List[VarDecl],
            pres: List[Expr], posts: List[Expr],
            local_vars: List[VarDecl], body: List[Stmt],
            position: Position, info: Info,
            method: PythonMethod = None,
            overriding_check: bool = False) -> Method:
        """Construct method AST node with additional obligation stuff."""
        return self._method_translator.create_method_node(
            ctx, name, args, returns, pres, posts, local_vars, body,
            position, info, method, overriding_check)

    def create_method_call_node(
            self, ctx: Context, method_name: str, args: List[Expr],
            targets: List[Expr], position: Position, info: Info,
            target_method: PythonMethod = None,
            target_node: ast.Call = None) -> List[Stmt]:
        """Construct a method call AST node with obligation stuff."""
        return self._method_translator.create_method_call_node(
            ctx, method_name, args, targets, position, info,
            target_method, target_node)

    def create_method_fork(self, ctx: Context, targets, thread: Expr,
                           position: Position, info: Info,
                           target_node: ast.Call = None) -> List[Stmt]:
        return self._method_translator.create_method_fork(
            ctx, targets, thread, position, info, target_node)

    def create_obligation_info(self, method: PythonMethod) -> object:
        """Create obligation info for the method."""
        info = PythonMethodObligationInfo(
            self._obligation_manager, method, self)
        info.traverse_contract()
        return info



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Classes for storing IO translation state."""


from collections import OrderedDict
from typing import List, Tuple

from nagini_translation.lib.program_nodes import PythonVar
from nagini_translation.lib.typedefs import Expr


class IOOpenContext:
    """Current state of IO operation open translation process."""

    def __init__(self) -> None:

        self._is_opening = False
        """Are we currently translating IO open?"""

        self._open_var_aliases = {}  # type: Dict[str, PythonVar]
        """Variables used while opening IO operation."""

        self._open_var_alias_definitions = OrderedDict()
        """Initial assignments to variables used in opening IO operation."""

    def start_io_operation_open(self) -> None:
        """Check state before IO operation open.

        This method is supposed to be called just before starting to
        translate IO operation open. It checks if context is in a good
        state.
        """
        assert not self._is_opening
        assert not self._open_var_aliases
        assert not self._open_var_alias_definitions
        self._is_opening = True

    def stop_io_operation_open(self) -> None:
        """Clean state after IO operation open.

        This method is supposed to be called just after finishing to
        translate IO operation open. It cleans up the context state.
        """
        assert self._is_opening
        self._is_opening = False
        self._open_var_aliases.clear()
        self._open_var_alias_definitions.clear()

    def add_variable(self, var_name: str, var: PythonVar) -> None:
        """Add IO opening variable."""
        assert self._is_opening
        self._open_var_aliases[var_name] = var

    def contains_variable(self, var_name: str) -> bool:
        """Check if variable is from IO opening."""
        return self._is_opening and var_name in self._open_var_aliases

    def get_variable(self, var_name) -> PythonVar:
        """Get variable that is used while opening IO operation."""
        assert self._is_opening
        return self._open_var_aliases[var_name]

    def is_variable_defined(self, var_name: str) -> bool:
        """Check if variable has already a definition."""
        assert self._is_opening
        return var_name in self._open_var_alias_definitions

    def define_variable(self, var_name: str, definition: Expr) -> None:
        """Assign definition to a variable."""
        assert self._is_opening
        self._open_var_alias_definitions[var_name] = definition

    def get_ordered_variable_defs(
            self) -> List[Tuple[PythonVar, Expr]]:
        """Get variables with their definitions.

        Returned variables are in their definition order.
        """
        assert self._is_opening
        result = []
        for name, definition in self._open_var_alias_definitions.items():
            result.append((self._open_var_aliases[name], definition))
        return result
